import pytest
import re
import datetime
from unittest import mock
from deepdiff.helper import number_to_string, CannotCompare
from deepdiff import DeepDiff
from decimal import Decimal
from deepdiff.deephash import sha256hex
from tests import CustomClass2


class TestIgnoreOrder:

    @pytest.mark.parametrize("t1, t2, significant_digits, ignore_order, result", [
        (10, 10.0, 5, False, {}),
        ({10: 'a', 11.1: 'b'}, {10.0: 'a', Decimal('11.1000003'): 'b'}, 5, False, {}),
    ])
    def test_type_change_numeric_ignored(self, t1, t2, significant_digits, ignore_order, result):
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True,
                         significant_digits=significant_digits, ignore_order=ignore_order)
        assert result == ddiff

    @pytest.mark.parametrize("t1, t2, expected_result",
                             [
                                 (10, 10.0, {}),
                                 (10, 10.2, {'values_changed': {'root': {'new_value': 10.2, 'old_value': 10}}}),
                                 (Decimal(10), 10.0, {}),
                                 ({"a": Decimal(10), "b": 12, 11.0: None}, {b"b": 12, "a": 10.0, Decimal(11): None}, {}),
                             ])
    def test_type_change_numeric_when_ignore_order(self, t1, t2, expected_result):
        ddiff = DeepDiff(t1, t2, ignore_order=True, ignore_numeric_type_changes=True, ignore_string_type_changes=True, threshold_to_diff_deeper=0)
        assert expected_result == ddiff

    def test_ignore_order_depth1(self):
        t1 = [{1, 2, 3}, {4, 5}]
        t2 = [{4, 5, 6}, {1, 2, 3}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {'set_item_added': ["root[1][6]"]} == ddiff

    def test_ignore_order_depth2(self):
        t1 = [[1, 2, 3], [4, 5]]
        t2 = [[4, 5, 6], [1, 2, 3]]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {'iterable_item_added': {'root[1][2]': 6}} == ddiff

    def test_ignore_order_depth3(self):
        t1 = [{1, 2, 3}, [{4, 5}]]
        t2 = [[{4, 5, 6}], {1, 2, 3}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {'set_item_added': ["root[1][0][6]"]} == ddiff
        assert {"root[1][0][6]"} == ddiff.affected_paths

    def test_ignore_order_depth4(self):
        t1 = [[1, 2, 3, 4], [4, 2, 2, 1]]
        t2 = [[4, 1, 1, 1], [1, 3, 2, 4]]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {'iterable_item_removed': {'root[1][1]': 2}} == ddiff

    def test_ignore_order_depth5(self):
        t1 = [4, 2, 2, 1]
        t2 = [4, 1, 1, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True, cache_purge_level=0)
        expected = {
            'iterable_item_removed': {
                'root[1]': 2,
                'root[2]': 2
            },
            'repetition_change': {
                'root[3]': {
                    'old_repeat': 1,
                    'new_repeat': 3,
                    'old_indexes': [3],
                    'new_indexes': [1, 2, 3],
                    'value': 1
                }
            }
        }
        assert expected == ddiff
        assert {"root[1]", "root[2]", "root[3]"} == ddiff.affected_paths

        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=False, cache_purge_level=0)
        dist = ddiff._get_rough_distance()
        assert 0.1 == dist

    def test_ignore_order_depth6(self):
        t1 = [[1, 2, 3, 4], [4, 2, 2, 1]]
        t2 = [[4, 1, 1, 1], [1, 3, 2, 4]]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        expected = {
            'iterable_item_removed': {
                'root[1][1]': 2,
                'root[1][2]': 2
            },
            'repetition_change': {
                'root[1][3]': {
                    'old_repeat': 1,
                    'new_repeat': 3,
                    'old_indexes': [3],
                    'new_indexes': [1, 2, 3],
                    'value': 1
                }
            }
        }

        assert expected == ddiff

    def test_list_difference_ignore_order(self):
        t1 = {1: 1, 4: {"a": "hello", "b": [1, 2, 3]}}
        t2 = {1: 1, 4: {"a": "hello", "b": [1, 3, 2, 3]}}
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    @pytest.mark.parametrize('t1_0, t2_0', [
        (1, 2),
        (True, False),
        ('a', 'b'),
    ])
    def test_list_difference_of_bool_only_ignore_order(self, t1_0, t2_0):
        t1 = [t1_0]
        t2 = [t2_0]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'values_changed': {'root[0]': {'new_value': t2_0, 'old_value': t1_0}}}
        assert result == ddiff

    def test_dictionary_difference_ignore_order(self):
        t1 = {"a": [[{"b": 2, "c": 4}, {"b": 2, "c": 3}]]}
        t2 = {"a": [[{"b": 2, "c": 3}, {"b": 2, "c": 4}]]}
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff
        assert set() == ddiff.affected_paths

    def test_nested_list_ignore_order(self):
        t1 = [1, 2, [3, 4]]
        t2 = [[4, 3, 3], 2, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_nested_list_difference_ignore_order(self):
        t1 = [1, 2, [3, 4]]
        t2 = [[4, 3], 2, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_nested_list_with_dictionarry_difference_ignore_order(self):
        t1 = [1, 2, [3, 4, {1: 2}]]
        t2 = [[4, 3, {1: 2}], 2, 1]

        ddiff = DeepDiff(t1, t2, ignore_order=True)

        result = {}
        assert result == ddiff

    def test_list_difference_ignore_order_report_repetition1(self):
        t1 = [1, 3, 1, 4]
        t2 = [4, 4, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        result = {
            'iterable_item_removed': {
                'root[1]': 3
            },
            'repetition_change': {
                'root[0]': {
                    'old_repeat': 2,
                    'old_indexes': [0, 2],
                    'new_indexes': [2],
                    'value': 1,
                    'new_repeat': 1
                },
                'root[3]': {
                    'old_repeat': 1,
                    'old_indexes': [3],
                    'new_indexes': [0, 1],
                    'value': 4,
                    'new_repeat': 2
                }
            }
        }
        assert result == ddiff

    @pytest.mark.skip
    def test_list_difference_ignore_order_report_repetition2(self):
        t1 = [1, 1, 1]
        t2 = [2, 2]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'values_changed': {'root[0]': {'new_value': 2, 'old_value': 1}}}
        assert result == ddiff

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, report_repetition=True, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1)
        result2 = {
            'iterable_item_removed': {
                'root[0]': 1,
                'root[1]': 1,
                'root[2]': 1
            },
            'iterable_item_added': {
                'root[0]': 2,
                'root[1]': 2,
            },
        }
        assert result2 == ddiff2

    @pytest.mark.skip
    def test_list_difference_ignore_order_report_repetition3(self):
        t1 = [{"id": 1}, {"id": 1}, {"id": 1}]
        t2 = [{"id": 1, "name": 1}]

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, report_repetition=True, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1)
        result2 = {
            'iterable_item_removed': {
                'root[1]': {"id": 1},
                'root[2]': {"id": 1},
            },
            'dictionary_item_added': ["root[0]['name']"]
        }
        assert result2 == ddiff2

    @pytest.mark.skip
    def test_list_difference_ignore_order_report_repetition4(self):
        t1 = [{"id": 1}, {"id": 1}, {"id": 1}, {"name": "Joe"}, {"name": "Joe"}]
        t2 = [{"id": 1, "name": 1}, {"id": 1, "name": "Joe"}]

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, report_repetition=True, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1)
        result2 = {
            'iterable_item_removed': {
                'root[2]': {"id": 1},
                'root[3]': {"name": "Joe"},
                'root[4]': {"name": "Joe"},
            },
            'dictionary_item_added': ["root[0]['name']", "root[1]['name']"]
        }
        assert result2 == ddiff2

    def test_nested_list_ignore_order_report_repetition(self):
        t1 = [1, 2, [3, 4]]
        t2 = [[4, 3, 3], 2, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=False)
        assert not ddiff

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        result = {
            'repetition_change': {
                'root[2][0]': {
                    'old_repeat': 1,
                    'new_repeat': 2,
                    'old_indexes': [0],
                    'new_indexes': [1, 2],
                    'value': 3
                }
            }
        }
        assert result == ddiff2
        assert {"root[2][0]"} == ddiff2.affected_paths

    @pytest.mark.skip
    def test_nested_list_and_dict_ignore_order_report_repetition(self):
        """
        This test shows that ignore order is not doing the right thing.

        It should have said that root[1] and root[2] are removed.
        """
        t1 = [{"id": 1}, {"id": 1}, {"id": 1}]
        t2 = [{"id": 1, "name": 1}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'dictionary_item_added': ["root[0]['name']"]}
        assert result == ddiff

        # Here there is nothing that is "repeated" in an iterable
        ddiff2 = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        assert result == ddiff2
        assert {"root[2][0]"} == ddiff2.affected_paths

    def test_list_of_unhashable_difference_ignore_order(self):
        t1 = [{"a": 2}, {"b": [3, 4, {1: 1}]}]
        t2 = [{"b": [3, 4, {1: 1}]}, {"a": 2}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_list_of_unhashable_difference_ignore_order2(self):
        t1 = [1, {"a": 2}, {"b": [3, 4, {1: 1}]}, "B"]
        t2 = [{"b": [3, 4, {1: 1}]}, {"a": 2}, {1: 1}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {
            'iterable_item_added': {
                'root[2]': {
                    1: 1
                }
            },
            'iterable_item_removed': {
                'root[3]': 'B',
                'root[0]': 1
            }
        }
        assert result == ddiff

    def test_list_of_unhashable_difference_ignore_order3(self):
        t1 = [1, {"a": 2}, {"a": 2}, {"b": [3, 4, {1: 1}]}, "B"]
        t2 = [{"b": [3, 4, {1: 1}]}, {1: 1}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {
            'values_changed': {
                'root[1]': {
                    'new_value': {
                        1: 1
                    },
                    'old_value': {
                        'a': 2
                    }
                }
            },
            'iterable_item_removed': {
                'root[0]': 1,
                'root[4]': 'B'
            }
        }
        assert result == ddiff

    def test_list_of_unhashable_difference_ignore_order_report_repetition(
            self):
        t1 = [1, {"a": 2}, {"a": 2}, {"b": [3, 4, {1: 1}]}, "B"]
        t2 = [{"b": [3, 4, {1: 1}]}, {1: 1}]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True, threshold_to_diff_deeper=0)
        result = {
            'iterable_item_added': {
                'root[1]': {
                    1: 1
                }
            },
            'iterable_item_removed': {
                'root[4]': 'B',
                'root[0]': 1,
                'root[1]': {
                    'a': 2
                },
                'root[2]': {
                    'a': 2
                }
            }
        }
        assert result == ddiff

    def test_list_of_unhashable_difference_ignore_order4(self):
        t1 = [{"a": 2}, {"a": 2}]
        t2 = [{"a": 2}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {}
        assert result == ddiff

    def test_list_of_unhashable_difference_ignore_order_report_repetition2(
            self):
        t1 = [{"a": 2}, {"a": 2}]
        t2 = [{"a": 2}]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        result = {
            'repetition_change': {
                'root[0]': {
                    'old_repeat': 2,
                    'new_indexes': [0],
                    'old_indexes': [0, 1],
                    'value': {
                        'a': 2
                    },
                    'new_repeat': 1
                }
            }
        }
        assert result == ddiff

    def test_list_ignore_order_report_repetition(self):
        t1 = [5, 1, 3, 1, 4, 4, 6]
        t2 = [7, 4, 4, 1, 3, 4, 8]
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        result = {
            'values_changed': {
                'root[6]': {
                    'new_value': 7,
                    'old_value': 6
                },
                'root[0]': {
                    'new_value': 8,
                    'old_value': 5
                }
            },
            'repetition_change': {
                'root[4]': {
                    'old_repeat': 2,
                    'new_repeat': 3,
                    'old_indexes': [4, 5],
                    'new_indexes': [1, 2, 5],
                    'value': 4
                },
                'root[1]': {
                    'old_repeat': 2,
                    'new_repeat': 1,
                    'old_indexes': [1, 3],
                    'new_indexes': [3],
                    'value': 1
                }
            }
        }
        assert result == ddiff

    def test_list_of_sets_difference_ignore_order(self):
        t1 = [{1}, {2}, {3}]
        t2 = [{4}, {1}, {2}, {3}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'iterable_item_added': {'root[0]': {4}}}
        assert result == ddiff

    def test_list_of_sets_difference_ignore_order_when_there_is_duplicate(
            self):
        t1 = [{1}, {2}, {3}]
        t2 = [{4}, {1}, {2}, {3}, {3}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'iterable_item_added': {'root[0]': {4}}}
        assert result == ddiff

    def test_list_of_sets_difference_ignore_order_when_there_is_duplicate_and_mix_of_hashable_unhashable(
            self):
        t1 = [1, 1, {2}, {3}]
        t2 = [{4}, 1, {2}, {3}, {3}, 1, 1]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {'iterable_item_added': {'root[0]': {4}}}
        assert result == ddiff

    def test_dictionary_of_list_of_dictionary_ignore_order(self):
        t1 = {
            'item': [{
                'title': 1,
                'http://purl.org/rss/1.0/modules/content/:encoded': '1'
            }, {
                'title': 2,
                'http://purl.org/rss/1.0/modules/content/:encoded': '2'
            }]
        }

        t2 = {
            'item': [{
                'http://purl.org/rss/1.0/modules/content/:encoded': '1',
                'title': 1
            }, {
                'http://purl.org/rss/1.0/modules/content/:encoded': '2',
                'title': 2
            }]
        }

        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_comprehensive_ignore_order(self):

        t1 = {
            'key1': 'val1',
            'key2': [
                {
                    'key3': 'val3',
                    'key4': 'val4',
                },
                {
                    'key5': 'val5',
                    'key6': 'val6',
                },
            ],
        }

        t2 = {
            'key1': 'val1',
            'key2': [
                {
                    'key5': 'val5',
                    'key6': 'val6',
                },
                {
                    'key3': 'val3',
                    'key4': 'val4',
                },
            ],
        }

        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_ignore_order_when_objects_similar(self):

        t1 = {
            'key1': 'val1',
            'key2': [
                {
                    'key3': 'val3',
                    'key4': 'val4',
                },
                {
                    'key5': 'val5',
                    'key6': 'val6',
                },
            ],
        }

        t2 = {
            'key1': 'val1',
            'key2': [
                {
                    'key5': 'CHANGE',
                    'key6': 'val6',
                },
                {
                    'key3': 'val3',
                    'key4': 'val4',
                },
            ],
        }

        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {'values_changed': {"root['key2'][1]['key5']": {'new_value': 'CHANGE', 'old_value': 'val5'}}} == ddiff

    def test_set_ignore_order_report_repetition(self):
        """
        If this test fails, it means that DeepDiff is not checking
        for set types before general iterables.
        So it forces creating the hashtable because of report_repetition=True.
        """
        t1 = {2, 1, 8}
        t2 = {1, 2, 3, 5}
        ddiff = DeepDiff(t1, t2, ignore_order=True, report_repetition=True)
        result = {
            'set_item_added': {'root[3]', 'root[5]'},
            'set_item_removed': {'root[8]'}
        }
        assert result == ddiff

    def test_custom_objects2(self):
        cc_a = CustomClass2(prop1=["a"], prop2=["b"])
        cc_b = CustomClass2(prop1=["b"], prop2=["b"])
        t1 = [cc_a, CustomClass2(prop1=["c"], prop2=["d"])]
        t2 = [cc_b, CustomClass2(prop1=["c"], prop2=["d"])]

        ddiff = DeepDiff(t1, t2, ignore_order=True)

        result = {'values_changed': {'root[0].prop1[0]': {'new_value': 'b', 'old_value': 'a'}}}
        assert result == ddiff

    def test_custom_object_type_change_when_ignore_order(self):

        class Burrito:
            bread = 'flour'

            def __init__(self):
                self.spicy = True

        class Taco:
            bread = 'flour'

            def __init__(self):
                self.spicy = True

        burrito = Burrito()
        taco = Taco()

        burritos = [burrito]
        tacos = [taco]

        assert not DeepDiff(burritos, tacos, ignore_type_in_groups=[(Taco, Burrito)], ignore_order=True)

    def test_decimal_ignore_order(self):
        t1 = [{1: Decimal('10.1')}, {2: Decimal('10.2')}]
        t2 = [{2: Decimal('10.2')}, {1: Decimal('10.1')}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        result = {}
        assert result == ddiff

    @pytest.mark.parametrize('log_scale_similarity_threshold, expected', [
        (
            0.1,
            {}
        ),
        (
            0.01,
            {'values_changed': {'root[1][2]': {'new_value': Decimal('268'), 'old_value': Decimal('290.2')}}}
        ),
    ])
    def test_decimal_log_scale_ignore_order1(self, log_scale_similarity_threshold, expected):
        t1 = [{1: Decimal('10.143')}, {2: Decimal('290.2')}]
        t2 = [{2: Decimal('268')}, {1: Decimal('10.23')}]
        ddiff = DeepDiff(t1, t2, ignore_order=True, use_log_scale=True, log_scale_similarity_threshold=log_scale_similarity_threshold, cutoff_intersection_for_pairs=1)
        assert expected == ddiff

    @pytest.mark.parametrize("t1, t2, significant_digits, ignore_order", [
        (100000, 100021, 3, False),
        ([10, 12, 100000], [50, 63, 100021], 3, False),
        ([10, 12, 100000], [50, 63, 100021], 3, True),
    ])
    def test_number_to_string_func(self, t1, t2, significant_digits, ignore_order):
        def custom_number_to_string(number, *args, **kwargs):
            number = 100 if number < 100 else number
            return number_to_string(number, *args, **kwargs)

        ddiff = DeepDiff(t1, t2, significant_digits=3, number_format_notation="e",
                         number_to_string_func=custom_number_to_string)

        assert {} == ddiff

    def test_ignore_type_in_groups_numbers_and_strings_when_ignore_order(self):
        t1 = [1, 2, 3, 'a']
        t2 = [1.0, 2.0, 3.3, b'a']
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True, ignore_string_type_changes=True, ignore_order=True)
        result = {'values_changed': {'root[2]': {'new_value': 3.3, 'old_value': 3}}}
        assert result == ddiff

    def test_ignore_string_type_changes_when_dict_keys_merge_is_not_deterministic(self):
        t1 = {'a': 10, b'a': 20}
        t2 = {'a': 11, b'a': 22}
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True, ignore_string_type_changes=True, ignore_order=True)
        result = {'values_changed': {"root['a']": {'new_value': 22, 'old_value': 20}}}
        alternative_result = {'values_changed': {"root['a']": {'new_value': 11, 'old_value': 10}}}
        assert result == ddiff or alternative_result == ddiff

    def test_skip_exclude_path5(self):
        exclude_paths = ["root[0]['e']", "root[1]['e']"]

        t1 = [{'a': 1, 'b': 'randomString', 'e': "1111"}]
        t2 = [{'a': 1, 'b': 'randomString', 'e': "2222"}]

        ddiff = DeepDiff(t1, t2, exclude_paths=exclude_paths,
                         ignore_order=True, report_repetition=False)
        result = {}
        assert result == ddiff

    def test_skip_str_type_in_dict_on_list_when_ignored_order(self):
        t1 = [{1: "a"}]
        t2 = [{}]
        ddiff = DeepDiff(t1, t2, exclude_types=[str], ignore_order=True)
        result = {}
        assert result == ddiff

    @mock.patch('deepdiff.diff.logger')
    @mock.patch('deepdiff.diff.DeepHash')
    def test_diff_when_hash_fails(self, mock_DeepHash, mock_logger):
        mock_DeepHash.side_effect = ValueError('Boom!')
        t1 = {"blah": {4}, 2: 1337}
        t2 = {"blah": {4}, 2: 1337}
        with pytest.raises(ValueError) as exp:
            DeepDiff(t1, t2, ignore_order=True)
        assert 'Boom!' == str(exp.value)

    def test_bool_vs_number(self):
        t1 = {
            "A List": [
                {
                    "Value One": True,
                    "Value Two": 1
                }
            ],
        }

        t2 = {
            "A List": [
                {
                    "Value Two": 1,
                    "Value One": True
                }
            ],
        }

        ddiff = DeepDiff(t1, t2, ignore_order=True, cutoff_intersection_for_pairs=1)
        assert ddiff == {}

    @pytest.mark.parametrize('max_passes, expected', [
        (0, {'values_changed': {'root[0]': {'new_value': {'key5': 'CHANGE', 'key6': 'val6'}, 'old_value': {'key3': [[[[[1, 2, 4, 5]]]]], 'key4': [7, 8]}}, 'root[1]': {'new_value': {'key3': [[[[[1, 3, 5, 4]]]]], 'key4': [7, 8]}, 'old_value': {'key5': 'val5', 'key6': 'val6'}}}}),
        (1, {'values_changed': {"root[1]['key5']": {'new_value': 'CHANGE', 'old_value': 'val5', 'new_path': "root[0]['key5']"}, "root[0]['key3'][0]": {'new_value': [[[[1, 3, 5, 4]]]], 'old_value': [[[[1, 2, 4, 5]]]], 'new_path': "root[1]['key3'][0]"}}}),
        (22, {'values_changed': {"root[1]['key5']": {'new_value': 'CHANGE', 'old_value': 'val5', 'new_path': "root[0]['key5']"}, "root[0]['key3'][0][0][0][0][1]": {'new_value': 3, 'old_value': 2, 'new_path': "root[1]['key3'][0][0][0][0][1]"}}})
    ])
    def test_ignore_order_max_passes(self, max_passes, expected):
        t1 = [
            {
                'key3': [[[[[1, 2, 4, 5]]]]],
                'key4': [7, 8],
            },
            {
                'key5': 'val5',
                'key6': 'val6',
            },
        ]

        t2 = [
            {
                'key5': 'CHANGE',
                'key6': 'val6',
            },
            {
                'key3': [[[[[1, 3, 5, 4]]]]],
                'key4': [7, 8],
            },
        ]

        ddiff = DeepDiff(t1, t2, ignore_order=True, max_passes=max_passes, verbose_level=2, cache_size=5000, cutoff_intersection_for_pairs=1, threshold_to_diff_deeper=0)
        assert expected == ddiff

    @pytest.mark.parametrize('max_diffs, expected', [
        (1, {}),
        (65, {'values_changed': {"root[1]['key5']": {'new_value': 'CHANGE', 'old_value': 'val5', 'new_path': "root[0]['key5']"}}}),
        (80, {'values_changed': {"root[1]['key5']": {'new_value': 'CHANGE', 'old_value': 'val5', 'new_path': "root[0]['key5']"}, "root[0]['key3'][0][0][0][0][1]": {'new_value': 3, 'old_value': 2, 'new_path': "root[1]['key3'][0][0][0][0][1]"}}}),
    ])
    def test_ignore_order_max_diffs(self, max_diffs, expected):
        t1 = [
            {
                'key3': [[[[[1, 2, 4, 5]]]]],
                'key4': [7, 8],
            },
            {
                'key5': 'val5',
                'key6': 'val6',
            },
        ]

        t2 = [
            {
                'key5': 'CHANGE',
                'key6': 'val6',
            },
            {
                'key3': [[[[[1, 3, 5, 4]]]]],
                'key4': [7, 8],
            },
        ]

        # Note: these tests are not exactly deterministic
        ddiff = DeepDiff(t1, t2, ignore_order=True, max_diffs=max_diffs, verbose_level=2, cache_size=5000, cutoff_intersection_for_pairs=1)
        assert expected == ddiff

    def test_stats_that_include_distance_cache_hits(self):
        t1 = [
            [1, 2, 3, 9], [9, 8, 5, 9]
        ]

        t2 = [
            [1, 2, 4, 10], [4, 2, 5]
        ]

        diff = DeepDiff(t1, t2, ignore_order=True, cache_size=5000, cutoff_intersection_for_pairs=1)
        expected = {
            'PASSES COUNT': 7,
            'DIFF COUNT': 37,
            'DISTANCE CACHE HIT COUNT': 0,
            'MAX PASS LIMIT REACHED': False,
            'MAX DIFF LIMIT REACHED': False,
        }
        assert expected == diff.get_stats()

    def test_ignore_order_report_repetition_and_self_loop(self):
        t1 = [[1, 2, 1, 3]]
        t1.append(t1)

        t2 = [[1, 2, 2, 2, 4]]
        t2.append(t2)

        diff = DeepDiff(t1, t2, ignore_order=True, cutoff_intersection_for_pairs=1)
        expected = {
            'values_changed': {
                'root[0][3]': {
                    'new_value': 4,
                    'old_value': 3
                },
                'root[1]': {
                    'new_value': t2,
                    'old_value': t1
                }
            }
        }
        assert expected == diff

        diff2 = DeepDiff(t1, t2, ignore_order=True, cache_size=5000, cutoff_intersection_for_pairs=1)
        assert expected == diff2

    def test_ignore_order_with_sha256_hash(self):
        t1 = [
            [1, 2, 3, 9], [9, 8, 5, 9]
        ]

        t2 = [
            [1, 2, 3, 10], [8, 2, 5]
        ]
        diff = DeepDiff(t1, t2, ignore_order=True, hasher=sha256hex, cutoff_intersection_for_pairs=1)
        expected = {
            'values_changed': {
                'root[0][3]': {
                    'new_value': 10,
                    'old_value': 9
                },
                'root[1][0]': {
                    'new_value': 2,
                    'old_value': 9
                }
            }
        }
        assert expected == diff

    def test_ignore_order_cache_for_individual_distances(self):
        t1 = [[1, 2, 'B', 3], 'B']
        t2 = [[1, 2, 3, 5], 5]
        diff = DeepDiff(t1, t2, ignore_order=True, cache_size=5000, cutoff_intersection_for_pairs=1)
        expected = {
            'values_changed': {
                'root[1]': {
                    'new_value': 5,
                    'old_value': 'B'
                }
            },
            'iterable_item_added': {
                'root[0][3]': 5
            },
            'iterable_item_removed': {
                'root[0][2]': 'B'
            }
        }
        assert expected == diff

        stats = diff.get_stats()
        expected_stats = {
            'PASSES COUNT': 3,
            'DIFF COUNT': 13,
            'DISTANCE CACHE HIT COUNT': 1,
            'MAX PASS LIMIT REACHED': False,
            'MAX DIFF LIMIT REACHED': False
        }
        assert expected_stats == stats

        t1 = [[1, 2, 'B', 3], 5]
        t2 = [[1, 2, 3, 5], 'B']
        diff2 = DeepDiff(t1, t2, ignore_order=True, cache_size=5000, cutoff_intersection_for_pairs=1)
        assert expected_stats == diff2.get_stats()

    def test_cutoff_distance_for_pairs(self):
        t1 = [[1.0]]
        t2 = [[20.0]]
        diff1 = DeepDiff(t1, t2, ignore_order=True, cutoff_distance_for_pairs=0.3)
        expected1 = {'values_changed': {'root[0][0]': {'new_value': 20.0, 'old_value': 1.0}}}
        assert expected1 == diff1

        diff2 = DeepDiff(t1, t2, ignore_order=True, cutoff_distance_for_pairs=0.1)
        expected2 = {'values_changed': {'root[0]': {'new_value': [20.0], 'old_value': [1.0]}}}
        assert expected2 == diff2

        diff_with_dist = DeepDiff(1.0, 20.0, get_deep_distance=True)
        expected = {'values_changed': {'root': {'new_value': 20.0, 'old_value': 1.0}}, 'deep_distance': 0.2714285714285714}

        assert expected == diff_with_dist

    def test_ignore_order_and_group_by1(self):
        t1 = [
            {'id': 'AA', 'name': 'Joe', 'ate': ['Nothing']},
            {'id': 'BB', 'name': 'James', 'ate': ['Chips', 'Cheese']},
            {'id': 'CC', 'name': 'Mike', 'ate': ['Apple']},
        ]

        t2 = [
            {'id': 'BB', 'name': 'James', 'ate': ['Chips', 'Brownies', 'Cheese']},
            {'id': 'AA', 'name': 'Joe', 'ate': ['Nothing']},
            {'id': 'CC', 'name': 'Mike', 'ate': ['Apple', 'Apple']},
        ]

        diff = DeepDiff(t1, t2, group_by='id', ignore_order=False)
        expected = {'iterable_item_added': {"root['BB']['ate'][1]": 'Brownies', "root['CC']['ate'][1]": 'Apple'}}
        assert expected == diff

        diff2 = DeepDiff(t1, t2, group_by='id', ignore_order=True)
        expected2 = {'iterable_item_added': {"root['BB']['ate'][1]": 'Brownies'}}
        assert expected2 == diff2

    def test_ignore_order_and_group_by2(self):
        t1_data = [{'id': '1', 'codes': ['1', '2', '3']}]
        t2_data = [{'id': '1', 'codes': ['1', '2', '4']}]
        diff = DeepDiff(t1_data, t2_data, group_by='id', ignore_order=True)
        expected = {'values_changed': {"root['1']['codes'][2]": {'new_value': '4', 'old_value': '3'}}}
        assert expected == diff

    def test_ignore_order_and_group_by3(self):
        t1 = [{
            'id':
            '5ec52e',
            'products': [{
                'lineNumber': 1,
                'productPrice': '2.39',
                'productQuantity': 2
            }, {
                'lineNumber': 2,
                'productPrice': '4.44',
                'productQuantity': 1
            }],
        }]

        t2 = [{
            'id':
            '5ec52e',
            'products': [
                {
                    'lineNumber': 2,
                    'productPrice': '4.44',
                    'productQuantity': 1
                },
                {
                    'lineNumber': 1,
                    'productPrice': '2.39',
                    'productQuantity': 2
                },
            ],
        }]

        diff = DeepDiff(t1, t2, group_by='id', ignore_order=True)
        assert {} == diff

    def test_ignore_order_and_group_by4(self):
        t1 = [
            {
                "id": "1",
                "field_01": {
                    "subfield_01": {
                        "subfield_02": {"subfield_03": "1"},
                    }
                },
            },
            {"id": "2", "field_01": ["1", "2", "3"]},
            {"id": "3", "field_01": ["1", "2", "3"]},
        ]
        t2 = [
            {
                "id": "1",
                "field_01": {
                    "subfield_01": {
                        "subfield_02": {"subfield_03": "2"},
                    }
                },
            },
            {"id": "2", "field_01": ["4", "5", "6"]},
            {"id": "3", "field_01": ["7", "8", "9"]},
        ]
        diff = DeepDiff(t1, t2, group_by='id', ignore_order=True)
        expected = {
            'values_changed': {
                "root['1']['field_01']['subfield_01']['subfield_02']['subfield_03']": {
                    'new_value': '2',
                    'old_value': '1'
                },
                "root['2']['field_01'][1]": {
                    'new_value': '5',
                    'old_value': '2'
                },
                "root['3']['field_01'][2]": {
                    'new_value': '9',
                    'old_value': '3'
                },
                "root['2']['field_01'][0]": {
                    'new_value': '4',
                    'old_value': '1'
                },
                "root['3']['field_01'][1]": {
                    'new_value': '8',
                    'old_value': '2'
                },
                "root['3']['field_01'][0]": {
                    'new_value': '7',
                    'old_value': '1'
                },
                "root['2']['field_01'][2]": {
                    'new_value': '6',
                    'old_value': '3'
                }
            }
        }

        assert expected == diff

    def test_math_epsilon_when_ignore_order_in_dictionary(self):
        a = {'x': 0.001}
        b = {'x': 0.0011}
        diff = DeepDiff(a, b, ignore_order=True)
        assert {'values_changed': {"root['x']": {'new_value': 0.0011, 'old_value': 0.001}}} == diff

        diff2 = DeepDiff(a, b, ignore_order=True, math_epsilon=0.01)
        assert {} == diff2

    def test_math_epsilon_when_ignore_order_in_list(self):
        a = [0.001, 2]
        b = [2, 0.0011]
        diff = DeepDiff(a, b, ignore_order=True)
        assert {'values_changed': {'root[0]': {'new_value': 0.0011, 'old_value': 0.001}}} == diff

        diff2 = DeepDiff(a, b, ignore_order=True, math_epsilon=0.01)
        assert {} == diff2

    def test_math_epsilon_when_ignore_order_in_nested_list(self):
        a = [{'x': 0.001}, {'y': 2.00002}]
        b = [{'x': 0.0011}, {'y': 2}]

        diff = DeepDiff(a, b, ignore_order=True, math_epsilon=0.01)
        expected = {'values_changed': {'root[0]': {'new_value': {'x': 0.0011}, 'old_value': {'x': 0.001}}, 'root[1]': {'new_value': {'y': 2}, 'old_value': {'y': 2.00002}}}}
        assert expected == diff

    def test_datetime_and_ignore_order(self):
        diff = DeepDiff(
            [{'due_date': datetime.date(2024, 2, 1)}],
            [{'due_date': datetime.date(2024, 2, 2)}],
            ignore_order=True,
            ignore_numeric_type_changes=True
        )
        assert {} != diff



class TestCompareFuncIgnoreOrder:

    def test_ignore_order_with_compare_func_to_guide_comparison(self):
        t1 = [
            {
                'id': 1,
                'value': [1]
            },
            {
                'id': 2,
                'value': [7, 8, 1]
            },
            {
                'id': 3,
                'value': [7, 8],
            },
        ]

        t2 = [
            {
                'id': 2,
                'value': [7, 8]
            },
            {
                'id': 3,
                'value': [7, 8, 1],
            },
            {
                'id': 1,
                'value': [1]
            },
        ]

        expected = {
            'values_changed': {
                "root[2]['id']": {
                    'new_value': 2,
                    'old_value': 3
                },
                "root[1]['id']": {
                    'new_value': 3,
                    'old_value': 2
                }
            }
        }

        expected_with_compare_func = {
            'iterable_item_added': {
                "root[2]['value'][2]": 1
            },
            'iterable_item_removed': {
                "root[1]['value'][2]": 1
            }
        }

        ddiff = DeepDiff(t1, t2, ignore_order=True)

        assert expected == ddiff

        def compare_func(x, y, level=None):
            try:
                return x['id'] == y['id']
            except Exception:
                raise CannotCompare() from None

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, iterable_compare_func=compare_func)
        assert expected_with_compare_func == ddiff2
        assert ddiff != ddiff2

        ddiff3 = DeepDiff(t1, t2, ignore_order=True, iterable_compare_func=compare_func, view='tree')
        assert 1 == ddiff3['iterable_item_removed'][0].t1
        assert 1 == ddiff3['iterable_item_added'][0].t2

    def test_ignore_order_with_compare_func_can_throw_cannot_compare(self):
        t1 = [
            {1},
            {
                'id': 2,
                'value': [7, 8, 1]
            },
            {
                'id': 3,
                'value': [7, 8],
            },
        ]

        t2 = [
            {
                'id': 2,
                'value': [7, 8]
            },
            {
                'id': 3,
                'value': [7, 8, 1],
            },
            {},
        ]

        expected = {
            'type_changes': {
                'root[0]': {
                    'old_type': set,
                    'new_type': dict,
                    'old_value': {1},
                    'new_value': {}
                }
            },
            'values_changed': {
                "root[2]['id']": {
                    'new_value': 2,
                    'old_value': 3
                },
                "root[1]['id']": {
                    'new_value': 3,
                    'old_value': 2
                }
            }
        }
        expected_with_compare_func = {
            'type_changes': {
                'root[0]': {
                    'old_type': set,
                    'new_type': dict,
                    'old_value': {1},
                    'new_value': {}
                }
            },
            'iterable_item_added': {
                "root[2]['value'][2]": 1
            },
            'iterable_item_removed': {
                "root[1]['value'][2]": 1
            }
        }

        ddiff = DeepDiff(t1, t2, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1, ignore_order=True, threshold_to_diff_deeper=0)
        assert expected == ddiff

        def compare_func(x, y, level=None):
            try:
                return x['id'] == y['id']
            except Exception:
                raise CannotCompare() from None

        ddiff2 = DeepDiff(t1, t2, ignore_order=True, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1, iterable_compare_func=compare_func, threshold_to_diff_deeper=0)
        assert expected_with_compare_func == ddiff2
        assert ddiff != ddiff2

    def test_ignore_order_with_compare_func_with_one_each_hashes_added_hashes_removed(self):
        """
        Scenario:
        In this example which demonstrates the problem... We have two dictionaries containing lists for
        individualNames.  Each list contains exactly 2 elements. The effective change is that we are
        replacing the 2nd element in the list.
        NOTE: This is considered a REPLACEMENT of the second element and not an UPDATE of the element
        because we are providing a custom compare_func which will determine matching elements based on
        the value of the nameIdentifier field.  If the custom compare_func is not used, then
        deepdiff.diff will mistakenly treat the difference as being individual field updates for every
        field in the second element of the list.

        Intent:
        Use our custom compare_func, since we have provided it.
        We need to fall into self._precalculate_distance_by_custom_compare_func
        To do this, we are proposing a change to deepdiff.diff line 1128:

        Original:
            if hashes_added and hashes_removed and self.iterable_compare_func and len(hashes_added) > 1 and len(hashes_removed) > 1:

        Proposed/Updated:
            if hashes_added and hashes_removed \
            and self.iterable_compare_func \
            and len(hashes_added) > 0 and len(hashes_removed) > 0:

        NOTE: It is worth mentioning that deepdiff.diff line 1121, might also benefit by changing the length conditions
        to evaluate for > 0 (rather than > 1).
        """

        t1 = {
            "individualNames": [
                {
                    "firstName": "Johnathan",
                    "lastName": "Doe",
                    "prefix": "COLONEL",
                    "middleName": "A",
                    "primaryIndicator": True,
                    "professionalDesignation": "PHD",
                    "suffix": "SR",
                    "nameIdentifier": "00001"
                },
                {
                    "firstName": "John",
                    "lastName": "Doe",
                    "prefix": "",
                    "middleName": "",
                    "primaryIndicator": False,
                    "professionalDesignation": "",
                    "suffix": "SR",
                    "nameIdentifier": "00002"
                }
            ]
        }

        t2 = {
            "individualNames": [
                {
                    "firstName": "Johnathan",
                    "lastName": "Doe",
                    "prefix": "COLONEL",
                    "middleName": "A",
                    "primaryIndicator": True,
                    "professionalDesignation": "PHD",
                    "suffix": "SR",
                    "nameIdentifier": "00001"
                },
                {
                    "firstName": "Johnny",
                    "lastName": "Doe",
                    "prefix": "",
                    "middleName": "A",
                    "primaryIndicator": False,
                    "professionalDesignation": "",
                    "suffix": "SR",
                    "nameIdentifier": "00003"
                }
            ]
        }
        def compare_func(item1, item2, level=None):
            print("*** inside compare ***")
            it1_keys = item1.keys()

            try:

                # --- individualNames ---
                if 'nameIdentifier' in it1_keys and 'lastName' in it1_keys:
                    match_result = item1['nameIdentifier'] == item2['nameIdentifier']
                    print("individualNames - matching result:", match_result)
                    return match_result
                else:
                    print("Unknown list item...", "matching result:", item1 == item2)
                    return item1 == item2
            except Exception:
                raise CannotCompare() from None
        # ---------------------------- End of nested function

        actual_diff = DeepDiff(t1, t2, report_repetition=True,
                             ignore_order=True, iterable_compare_func=compare_func, cutoff_intersection_for_pairs=1)

        old_invalid_diff = {
            'values_changed': {"root['individualNames'][1]['firstName']": {'new_value': 'Johnny', 'old_value': 'John'},
                               "root['individualNames'][1]['middleName']": {'new_value': 'A', 'old_value': ''},
                               "root['individualNames'][1]['nameIdentifier']": {'new_value': '00003',
                                                                                'old_value': '00002'}}}
        new_expected_diff = {'iterable_item_added': {
            "root['individualNames'][1]": {'firstName': 'Johnny', 'lastName': 'Doe', 'prefix': '', 'middleName': 'A',
                                           'primaryIndicator': False, 'professionalDesignation': '', 'suffix': 'SR',
                                           'nameIdentifier': '00003'}}, 'iterable_item_removed': {
            "root['individualNames'][1]": {'firstName': 'John', 'lastName': 'Doe', 'prefix': '', 'middleName': '',
                                           'primaryIndicator': False, 'professionalDesignation': '', 'suffix': 'SR',
                                           'nameIdentifier': '00002'}}}

        assert old_invalid_diff != actual_diff
        assert new_expected_diff == actual_diff


class TestDynamicIgnoreOrder:
    def test_ignore_order_func(self):
        t1 = {
            "order_matters": [
                {1},
                {
                    'id': 2,
                    'value': [7, 8, 1]
                },
                {
                    'id': 3,
                    'value': [7, 8],
                },
            ],
            "order_does_not_matter": [
                {1},
                {
                    'id': 2,
                    'value': [7, 8, 1]
                },
                {
                    'id': 3,
                    'value': [7, 8],
                },
            ]
        }

        t2 = {
            "order_matters": [
                {
                    'id': 2,
                    'value': [7, 8]
                },
                {
                    'id': 3,
                    'value': [7, 8, 1],
                },
                {},
            ],
            "order_does_not_matter": [
                {
                    'id': 2,
                    'value': [7, 8]
                },
                {
                    'id': 3,
                    'value': [7, 8, 1],
                },
                {},
            ]
        }

        def ignore_order_func(level):
            return "order_does_not_matter" in level.path()

        ddiff = DeepDiff(t1, t2, cutoff_intersection_for_pairs=1, cutoff_distance_for_pairs=1, ignore_order_func=ignore_order_func, threshold_to_diff_deeper=0)

        expected = {
            'type_changes': {
                "root['order_matters'][0]": {
                    'old_type': set,
                    'new_type': dict,
                    'old_value': {1},
                    'new_value': {'id': 2, 'value': [7, 8]}
                },
                "root['order_does_not_matter'][0]": {
                    'old_type': set,
                    'new_type': dict,
                    'old_value': {1},
                    'new_value': {}
                }
            },
            'dictionary_item_removed': [
                "root['order_matters'][2]['id']",
                "root['order_matters'][2]['value']"
            ],
            'values_changed': {
                "root['order_matters'][1]['id']": {'new_value': 3, 'old_value': 2},
                "root['order_does_not_matter'][2]['id']": {'new_value': 2, 'old_value': 3},
                "root['order_does_not_matter'][1]['id']": {'new_value': 3, 'old_value': 2}
            }
        }
        assert expected == ddiff


class TestDecodingErrorIgnoreOrder:

    EXPECTED_MESSAGE1 = (
        "'utf-8' codec can't decode byte 0xc3 in position 0: Can not produce a hash for root: invalid continuation byte in '('. "
        "Please either pass ignore_encoding_errors=True or pass the encoding via encodings=['utf-8', '...'].")

    EXPECTED_MESSAGE2 = (
        "'utf-8' codec can't decode byte 0xbc in position 0: Can not produce a hash for root: invalid start byte in ' cup of flour'. "
        "Please either pass ignore_encoding_errors=True or pass the encoding via encodings=['utf-8', '...'].")

    @pytest.mark.parametrize('test_num, item, encodings, ignore_encoding_errors, expected_result, expected_message', [
        (1, b'\xc3\x28', None, False, UnicodeDecodeError, EXPECTED_MESSAGE1),
        (2, b'\xc3\x28', ['utf-8'], False, UnicodeDecodeError, EXPECTED_MESSAGE1),
        (3, b'\xc3\x28', ['utf-8'], True, {'values_changed': {'root[0]': {'new_value': b'\xc3(', 'old_value': b'foo'}}}, None),
        (4, b"\xbc cup of flour", ['utf-8'], False, UnicodeDecodeError, EXPECTED_MESSAGE2),
        (5, b"\xbc cup of flour", ['utf-8'], True, {'values_changed': {'root[0]': {'new_value': b'\xbc cup of flour', 'old_value': b'foo'}}}, None),
        (6, b"\xbc cup of flour", ['utf-8', 'latin-1'], False, {'values_changed': {'root[0]': {'new_value': b'\xbc cup of flour', 'old_value': b'foo'}}}, None),
    ])
    @mock.patch('deepdiff.diff.logger')
    def test_diff_encodings(self, mock_logger, test_num, item, encodings, ignore_encoding_errors, expected_result, expected_message):
        if UnicodeDecodeError == expected_result:
            with pytest.raises(expected_result) as exc_info:
                DeepDiff([b'foo'], [item], encodings=encodings, ignore_encoding_errors=ignore_encoding_errors, ignore_order=True)
            assert expected_message == str(exc_info.value), f"test_diff_encodings test #{test_num} failed."
        else:
            result = DeepDiff([b'foo'], [item], encodings=encodings, ignore_encoding_errors=ignore_encoding_errors, ignore_order=True)
            assert expected_result == result, f"test_diff_encodings test #{test_num} failed."


class TestErrorMessagesWhenIgnoreOrder:

    @mock.patch('deepdiff.diff.logger')
    def test_error_messages_when_ignore_order(self, mock_logger):
        t1 = {'x': 0, 'y': [0, 'a', 'b', 'c']}
        t2 = {'x': 1, 'y': [1, 'c', 'b', 'a']}

        exclude = [re.compile(r"\['x'\]"), re.compile(r"\['y'\]\[0\]")]

        result = DeepDiff(t1, t2, ignore_order=True, exclude_regex_paths=exclude)
        assert {} == result

        assert not mock_logger.error.called



class TestIgnoreOrderNumericTypeChange:
    """Regression tests for GitHub issue #485.

    When ignore_order=True, numerically-equal values of different numeric types
    (e.g. int 1 vs float 1.0) must still be reported as type_changes.

    Root cause: Python's hash equality (hash(1) == hash(1.0)) and value equality
    (1 == 1.0) caused both items to land in the same DeepHash bucket, so they
    were treated as identical and silently dropped from the diff result.
    """

    def test_int_vs_float_in_list_of_dicts(self):
        """Core regression: type change inside a dict nested in a list."""
        result = DeepDiff([{"a": 1}], [{"a": 1.0}], ignore_order=True)
        assert "type_changes" in result, (
            "Expected type_changes between int 1 and float 1.0, got: %s" % result
        )
        assert result["type_changes"]["root[0]['a']"]["old_type"] is int
        assert result["type_changes"]["root[0]['a']"]["new_type"] is float

    def test_ignore_numeric_type_changes_suppresses_report(self):
        """When ignore_numeric_type_changes=True the type change must be hidden."""
        result = DeepDiff(
            [{"a": 1}], [{"a": 1.0}],
            ignore_order=True,
            ignore_numeric_type_changes=True,
        )
        assert result == {}, (
            "With ignore_numeric_type_changes=True there should be no diff, got: %s" % result
        )

    def test_value_change_still_detected(self):
        """Ordinary value differences must still be detected."""
        result = DeepDiff([1], [2], ignore_order=True)
        assert result != {}, "Expected a diff between [1] and [2]"

    def test_reorder_no_false_positive(self):
        """A simple reorder of identical values must not trigger type_changes."""
        result = DeepDiff([1, 2, 3], [3, 2, 1], ignore_order=True)
        assert result == {}, "Reordering identical ints must not produce a diff"

    def test_mixed_list_one_type_change(self):
        """Only the item with a type change should appear in the diff."""
        result = DeepDiff(
            [{"a": 1}, {"b": 2}],
            [{"b": 2}, {"a": 1.0}],
            ignore_order=True,
        )
        assert "type_changes" in result
        assert "root[0]['a']" in result["type_changes"]

    def test_ignore_order_false_unchanged(self):
        """The ignore_order=False path must continue to work as before."""
        result = DeepDiff([{"a": 1}], [{"a": 1.0}], ignore_order=False)
        assert "type_changes" in result



#!/usr/bin/env python3
#
# DeepDiff documentation build configuration file, created by
# sphinx-quickstart on Mon Jul 20 06:06:44 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import datetime
from dotenv import load_dotenv

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# import pdb; pdb.set_trace()
sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx_sitemap',
    'sphinxemoji.sphinxemoji',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.

year = datetime.datetime.now().year
project = 'DeepDiff'
copyright = '2015-{}, Sep Dehpour'.format(year)
author = 'Sep Dehpour'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '9.0.0'
# The full version, including alpha/beta/rc tags.
release = '9.0.0'

load_dotenv(override=True)
DOC_VERSION = os.environ.get('DOC_VERSION', version)


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# Let Furo use its default pygments styles (light + dark mode aware)
pygments_style = 'xcode'
pygments_dark_style = 'github-dark'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'

html_theme_options = {
    'footer_icons': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/seperman/deepdiff',
            'html': '<svg stroke="currentColor" fill="currentColor" stroke-width="0" '
                    'viewBox="0 0 16 16"><path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 '
                    '3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-'
                    '2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 '
                    '1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-'
                    '3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 '
                    '.67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-'
                    '.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 '
                    '3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46'
                    '.55.38A8.013 8.013 0 0016 8c0-4.42-3.58-8-8-8z"></path></svg>',
            'class': '',
        },
    ],
    'source_repository': 'https://github.com/seperman/deepdiff',
    'source_branch': 'master',
    'source_directory': 'docs/',
    'light_css_variables': {
        # Fonts
        'font-stack': "'Open Sans', sans-serif",
        'font-stack--headings': "'Open Sans', sans-serif",
        'font-size--normal': '17px',
        # Body colors (from Alabaster)
        'color-foreground-primary': '#3E4349',
        'color-foreground-secondary': '#555',
        'color-foreground-muted': '#888',
        'color-background-primary': '#fff',
        'color-background-secondary': '#f8f8f8',
        'color-background-border': '#EEE',
        # Links
        'color-link': '#004B6B',
        'color-link--hover': '#6D4100',
        # Sidebar
        'color-sidebar-background': '#fff',
        'color-sidebar-text': '#555',
        'color-sidebar-link': '#444',
        'color-sidebar-link-text': '#444',
        'color-sidebar-link-text--top-level': '#444',
        'color-sidebar-item-background--hover': '#EEE',
        'color-sidebar-item-expander-color': '#444',
        'color-sidebar-caption-text': '#444',
        'color-sidebar-item-background--current': 'transparent',
        'color-sidebar-search-border': '#CCC',
        # Code blocks
        'color-code-background': '#ecf0f3',
        'color-code-foreground': '#222',
        'color-inline-code-background': '#ecf0f3',
        # Admonitions
        'color-admonition-background': '#EEE',
        'color-admonition-title': '#195190',
        'color-admonition-title-background': 'rgba(25, 81, 144, 0.1)',
        # Table of contents (right sidebar)
        'color-toc-item-text': '#444',
        'color-toc-item-text--hover': '#195190',
        'color-toc-item-text--active': '#195190',
        # Headerlink
        'color-header-text': '#3E4349',
    },
    'dark_css_variables': {
        'font-stack': "'Open Sans', sans-serif",
        'font-stack--headings': "'Open Sans', sans-serif",
        'font-size--normal': '17px',
        # Links
        'color-link': '#7EB8DA',
        'color-link--hover': '#D4A76A',
        # Sidebar
        'color-sidebar-background': '#1a1c1e',
        'color-sidebar-background-border': '#2d2d2d',
        'color-sidebar-link-text': '#ccc',
        'color-sidebar-link-text--top-level': '#ccc',
        'color-sidebar-caption-text': '#ccc',
        'color-sidebar-item-background--hover': '#2d2d2d',
        # Code blocks
        'color-code-background': '#2b2b2b',
        'color-inline-code-background': '#2b2b2b',
        # Admonitions
        'color-admonition-title': '#7EB8DA',
        'color-admonition-title-background': 'rgba(126, 184, 218, 0.15)',
        # TOC
        'color-toc-item-text--hover': '#7EB8DA',
        'color-toc-item-text--active': '#7EB8DA',
    },
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/qluster_grey_on_white_bg_optimized.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "./_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['custom.css']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr'
html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'DeepDiffdoc'

# Sitemap settings:

html_baseurl = f'https://zepworks.com/deepdiff/{DOC_VERSION}'
sitemap_url_scheme = "{link}"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'DeepDiff.tex', 'DeepDiff Documentation',
   'Sep Dehpour', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'deepdiff', 'DeepDiff Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'DeepDiff', 'DeepDiff Documentation',
   author, 'DeepDiff', 'Deep difference of Python objects.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False



import json
import os
from ast import literal_eval
from importlib.util import find_spec
from typing import Any, Dict

from deepdiff.model import TextResult, TreeResult


if os.name == "nt" and find_spec("colorama"):
    import colorama

    colorama.init()


# ANSI color codes
RED = '\033[31m'
GREEN = '\033[32m'
RESET = '\033[0m'


class ColoredView:
    """A view that shows JSON with color-coded differences."""

    def __init__(self, t2: Any, tree_result: TreeResult, compact: bool = False):
        self.t2 = t2
        self.tree = tree_result
        self.compact = compact
        self.diff_paths = self._collect_diff_paths()

    def _collect_diff_paths(self) -> Dict[str, str]:
        """Collect all paths that have differences and their types."""
        text_result = TextResult(tree_results=self.tree, verbose_level=2)
        diff_paths = {}
        for diff_type, items in text_result.items():
            if not items:
                continue
            try:
                iter(items)
            except TypeError:
                continue
            for path, item in items.items():
                if diff_type in ("values_changed", "type_changes"):
                    changed_path = item.get("new_path") or path
                    diff_paths[changed_path] = ("changed", item["old_value"], item["new_value"])
                elif diff_type in ("dictionary_item_added", "iterable_item_added", "set_item_added"):
                    diff_paths[path] = ("added", None, item)
                elif diff_type in ("dictionary_item_removed", "iterable_item_removed", "set_item_removed"):
                    diff_paths[path] = ("removed", item, None)
        return diff_paths

    def _format_value(self, value: Any) -> str:
        """Format a value for display."""
        if isinstance(value, bool):
            return 'true' if value else 'false'
        elif isinstance(value, str):
            return f'"{value}"'
        elif isinstance(value, (dict, list, tuple)):
            return json.dumps(value)
        else:
            return str(value)

    def _get_path_removed(self, path: str) -> dict:
        """Get all removed items for a given path."""
        removed = {}
        for key, value in self.diff_paths.items():
            if value[0] == 'removed' and key.startswith(path + "["):
                key_suffix = key[len(path):]
                if key_suffix.count("[") == 1 and key_suffix.endswith("]"):
                    removed[literal_eval(key_suffix[1:-1])] = value[1]
        return removed

    def _has_differences(self, path_prefix: str) -> bool:
        """Check if a path prefix has any differences under it."""
        return any(diff_path.startswith(path_prefix + "[") for diff_path in self.diff_paths)

    def _colorize_json(self, obj: Any, path: str = 'root', indent: int = 0) -> str:
        """Recursively colorize JSON based on differences, with pretty-printing."""
        INDENT = '  '
        current_indent = INDENT * indent
        next_indent = INDENT * (indent + 1)

        if path in self.diff_paths and path not in self._colorize_skip_paths:
            diff_type, old, new = self.diff_paths[path]
            if diff_type == 'changed':
                return f"{RED}{self._format_value(old)}{RESET} -> {GREEN}{self._format_value(new)}{RESET}"
            elif diff_type == 'added':
                return f"{GREEN}{self._format_value(new)}{RESET}"
            elif diff_type == 'removed':
                return f"{RED}{self._format_value(old)}{RESET}"

        if isinstance(obj, (dict, list)) and self.compact and not self._has_differences(path):
            return '{...}' if isinstance(obj, dict) else '[...]'

        if isinstance(obj, dict):
            if not obj:
                return '{}'
            items = []
            for key, value in obj.items():
                new_path = f"{path}['{key}']" if isinstance(key, str) else f"{path}[{key}]"
                if new_path in self.diff_paths and self.diff_paths[new_path][0] == 'added':
                    # Colorize both key and value for added fields
                    items.append(f'{next_indent}{GREEN}"{key}": {self._colorize_json(value, new_path, indent + 1)}{RESET}')
                else:
                    items.append(f'{next_indent}"{key}": {self._colorize_json(value, new_path, indent + 1)}')
            for key, value in self._get_path_removed(path).items():
                new_path = f"{path}['{key}']" if isinstance(key, str) else f"{path}[{key}]"
                items.append(f'{next_indent}{RED}"{key}": {self._colorize_json(value, new_path, indent + 1)}{RESET}')
            return '{\n' + ',\n'.join(items) + f'\n{current_indent}' + '}'

        elif isinstance(obj, (list, tuple)):
            removed_map = self._get_path_removed(path)
            if not obj and not removed_map:
                return '[]'

            for index in removed_map:
                self._colorize_skip_paths.add(f"{path}[{index}]")

            items = []
            remove_index = 0
            for index, value in enumerate(obj):
                while remove_index == next(iter(removed_map), None):
                    items.append(f'{next_indent}{RED}{self._format_value(removed_map.pop(remove_index))}{RESET}')
                    remove_index += 1
                items.append(f'{next_indent}{self._colorize_json(value, f"{path}[{index}]", indent + 1)}')
                remove_index += 1
            for value in removed_map.values():
                items.append(f'{next_indent}{RED}{self._format_value(value)}{RESET}')
            return '[\n' + ',\n'.join(items) + f'\n{current_indent}' + ']'
        else:
            return self._format_value(obj)

    def __str__(self) -> str:
        """Return the colorized, pretty-printed JSON string."""
        self._colorize_skip_paths = set()
        return self._colorize_json(self.t2)

    def __iter__(self):
        """Make the view iterable by yielding the tree results."""
        yield from self.tree.items()



#!/usr/bin/env python
"""Tests for fractions.Fraction support in DeepDiff."""
import pytest
import logging
from fractions import Fraction
from decimal import Decimal
from functools import partial
from deepdiff import DeepDiff, DeepHash
from deepdiff.deephash import prepare_string_for_hashing
from deepdiff.helper import number_to_string
from deepdiff.serialization import json_dumps, json_loads


logging.disable(logging.CRITICAL)

# Only the prep part of DeepHash. We don't need to test the actual hash function.
DeepHashPrep = partial(DeepHash, apply_hash=False)


class TestFractionDiff:
    """Tests for DeepDiff with Fraction objects."""

    def test_fraction_value_change(self):
        t1 = {1: Fraction(1, 3)}
        t2 = {1: Fraction(2, 3)}
        ddiff = DeepDiff(t1, t2)
        result = {
            'values_changed': {
                'root[1]': {
                    'new_value': Fraction(2, 3),
                    'old_value': Fraction(1, 3)
                }
            }
        }
        assert result == ddiff

    def test_fraction_no_change(self):
        t1 = Fraction(1, 3)
        t2 = Fraction(1, 3)
        ddiff = DeepDiff(t1, t2)
        assert {} == ddiff

    def test_fraction_vs_float_type_change(self):
        t1 = Fraction(1, 2)
        t2 = 0.5
        ddiff = DeepDiff(t1, t2)
        assert 'type_changes' in ddiff
        assert ddiff['type_changes']['root']['old_type'] == Fraction
        assert ddiff['type_changes']['root']['new_type'] == float

    def test_fraction_vs_int_type_change(self):
        t1 = Fraction(2, 1)
        t2 = 2
        ddiff = DeepDiff(t1, t2)
        assert 'type_changes' in ddiff
        assert ddiff['type_changes']['root']['old_type'] == Fraction
        assert ddiff['type_changes']['root']['new_type'] == int

    def test_fraction_vs_decimal_type_change(self):
        t1 = Fraction(1, 2)
        t2 = Decimal('0.5')
        ddiff = DeepDiff(t1, t2)
        assert 'type_changes' in ddiff

    def test_fraction_in_dict(self):
        t1 = {"a": Fraction(1, 3), "b": Fraction(2, 3)}
        t2 = {"a": Fraction(1, 3), "b": Fraction(3, 4)}
        ddiff = DeepDiff(t1, t2)
        assert 'values_changed' in ddiff
        assert "root['b']" in ddiff['values_changed']

    def test_fraction_in_list(self):
        t1 = [Fraction(1, 2), Fraction(1, 3)]
        t2 = [Fraction(1, 2), Fraction(1, 4)]
        ddiff = DeepDiff(t1, t2)
        result = {
            'values_changed': {
                'root[1]': {
                    'new_value': Fraction(1, 4),
                    'old_value': Fraction(1, 3)
                }
            }
        }
        assert result == ddiff

    def test_fraction_nested(self):
        t1 = {"data": [{"val": Fraction(1, 3)}]}
        t2 = {"data": [{"val": Fraction(2, 3)}]}
        ddiff = DeepDiff(t1, t2)
        assert 'values_changed' in ddiff
        assert "root['data'][0]['val']" in ddiff['values_changed']


class TestFractionIgnoreNumericTypeChanges:
    """Tests for ignore_numeric_type_changes with Fraction."""

    def test_fraction_vs_float_ignored(self):
        t1 = Fraction(1, 2)
        t2 = 0.5
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff

    def test_fraction_vs_int_ignored(self):
        t1 = Fraction(2, 1)
        t2 = 2
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff

    def test_fraction_vs_decimal_ignored(self):
        t1 = Fraction(1, 2)
        t2 = Decimal('0.5')
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff

    def test_fraction_vs_float_different_values(self):
        t1 = Fraction(1, 3)
        t2 = 0.5
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert 'values_changed' in ddiff

    def test_fraction_vs_float_in_list_ignored(self):
        t1 = [Fraction(1, 2), Fraction(3, 4)]
        t2 = [0.5, 0.75]
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff

    def test_fraction_vs_int_in_dict_ignored(self):
        t1 = {"a": Fraction(5, 1), "b": Fraction(10, 1)}
        t2 = {"a": 5, "b": 10}
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff

    @pytest.mark.parametrize("t1, t2, significant_digits, result", [
        ([0.5], [Fraction(1, 2)], 5, {}),
        ([Fraction(1, 3)], [0.333333], 5, {}),
        ([Fraction(1, 3)], [Decimal('0.33333')], 5, {}),
        ([1], [Fraction(1, 1)], 5, {}),
        ([-Fraction(1, 2)], [-0.5], 5, {}),
        ([Fraction(22, 7)], [3.14286], 4, {}),
    ])
    def test_ignore_numeric_type_changes_with_fraction(self, t1, t2, significant_digits, result):
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True, significant_digits=significant_digits)
        assert result == ddiff


class TestFractionSignificantDigits:
    """Tests for significant_digits with Fraction."""

    def test_fraction_significant_digits_equal(self):
        t1 = Fraction(1, 3)  # 0.333...
        t2 = Fraction(334, 1000)  # 0.334
        ddiff = DeepDiff(t1, t2, significant_digits=2)
        assert {} == ddiff

    def test_fraction_significant_digits_different(self):
        t1 = Fraction(1, 3)  # 0.333...
        t2 = Fraction(1, 2)  # 0.5
        ddiff = DeepDiff(t1, t2, significant_digits=1)
        assert 'values_changed' in ddiff

    @pytest.mark.parametrize("test_num, t1, t2, significant_digits, number_format_notation, result", [
        (1, Fraction(1, 3), Fraction(334, 1000), 2, "f", {}),
        (2, Fraction(1, 2), Fraction(499, 1000), 2, "f", {}),
        (3, Fraction(1, 2), Fraction(1, 3), 0, "f", {}),
        (4, Fraction(1, 2), Fraction(1, 3), 1, "f",
            {'values_changed': {'root': {'new_value': Fraction(1, 3), 'old_value': Fraction(1, 2)}}}),
        (5, Fraction(22, 7), Fraction(355, 113), 2, "f", {}),  # Two approximations of pi agree to 2 digits
        (6, Fraction(22, 7), Fraction(355, 113), 3, "f",
            {'values_changed': {'root': {'new_value': Fraction(355, 113), 'old_value': Fraction(22, 7)}}}),
    ])
    def test_fraction_significant_digits_and_notation(self, test_num, t1, t2, significant_digits, number_format_notation, result):
        ddiff = DeepDiff(t1, t2, significant_digits=significant_digits,
                         number_format_notation=number_format_notation)
        assert result == ddiff, f"test_fraction_significant_digits_and_notation #{test_num} failed."


class TestFractionMathEpsilon:
    """Tests for math_epsilon with Fraction."""

    def test_fraction_math_epsilon_close(self):
        d1 = {"a": Fraction(7175, 1000)}
        d2 = {"a": Fraction(7174, 1000)}
        res = DeepDiff(d1, d2, math_epsilon=0.01)
        assert res == {}

    def test_fraction_math_epsilon_not_close(self):
        d1 = {"a": Fraction(7175, 1000)}
        d2 = {"a": Fraction(7174, 1000)}
        res = DeepDiff(d1, d2, math_epsilon=0.0001)
        assert 'values_changed' in res

    def test_fraction_vs_float_math_epsilon(self):
        d1 = {"a": Fraction(1, 3)}
        d2 = {"a": 0.333}
        res = DeepDiff(d1, d2, math_epsilon=0.001, ignore_numeric_type_changes=True)
        assert res == {}


class TestFractionIgnoreOrder:
    """Tests for ignore_order with Fraction."""

    def test_fraction_ignore_order(self):
        t1 = [{1: Fraction(1, 3)}, {2: Fraction(2, 3)}]
        t2 = [{2: Fraction(2, 3)}, {1: Fraction(1, 3)}]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert {} == ddiff

    def test_fraction_ignore_order_with_change(self):
        t1 = [Fraction(1, 2), Fraction(1, 3)]
        t2 = [Fraction(1, 3), Fraction(1, 4)]
        ddiff = DeepDiff(t1, t2, ignore_order=True)
        assert ddiff != {}


class TestFractionAsKey:
    """Tests for Fraction used as dictionary key."""

    def test_fraction_as_dict_key(self):
        t1 = {Fraction(1, 2): "half"}
        t2 = {Fraction(1, 2): "one half"}
        ddiff = DeepDiff(t1, t2)
        assert 'values_changed' in ddiff

    def test_fraction_vs_float_key(self):
        # Fraction(1, 2) == 0.5 and hash(Fraction(1, 2)) == hash(0.5) in Python,
        # so they resolve to the same dict key. DeepDiff sees no difference.
        t1 = {Fraction(1, 2): "value"}
        t2 = {0.5: "value"}
        ddiff = DeepDiff(t1, t2)
        assert ddiff == {}

    def test_fraction_vs_float_key_ignore_numeric(self):
        t1 = {Fraction(1, 2): "value"}
        t2 = {0.5: "value"}
        ddiff = DeepDiff(t1, t2, ignore_numeric_type_changes=True)
        assert {} == ddiff


class TestFractionNumberToString:
    """Tests for number_to_string with Fraction."""

    @pytest.mark.parametrize("t1, t2, significant_digits, number_format_notation, expected_result", [
        (Fraction(1, 3), 0.333333, 5, "f", True),
        (Fraction(1, 2), 0.5, 5, "f", True),
        (Fraction(1, 2), 0.5, 5, "e", True),
        (Fraction(1, 3), Fraction(1, 4), 1, "f", ('0.3', '0.2')),
        (Fraction(22, 7), 3.14286, 4, "f", True),
        (Fraction(0), 0.0, 5, "f", True),
        (Fraction(-1, 2), -0.5, 5, "f", True),
    ])
    def test_number_to_string_fraction(self, t1, t2, significant_digits, number_format_notation, expected_result):
        st1 = number_to_string(t1, significant_digits=significant_digits, number_format_notation=number_format_notation)
        st2 = number_to_string(t2, significant_digits=significant_digits, number_format_notation=number_format_notation)
        if expected_result is True:
            assert st1 == st2
        else:
            assert st1 == expected_result[0]
            assert st2 == expected_result[1]


class TestFractionDeepHash:
    """Tests for DeepHash with Fraction."""

    def test_fraction_hash(self):
        result = DeepHash(Fraction(1, 3))
        assert result[Fraction(1, 3)]

    def test_fraction_same_value_same_hash(self):
        result1 = DeepHash(Fraction(1, 2))
        result2 = DeepHash(Fraction(1, 2))
        assert result1[Fraction(1, 2)] == result2[Fraction(1, 2)]

    def test_fraction_different_value_different_hash(self):
        result1 = DeepHash(Fraction(1, 2))
        result2 = DeepHash(Fraction(1, 3))
        assert result1[Fraction(1, 2)] != result2[Fraction(1, 3)]

    def test_fraction_vs_float_hash_different_by_default(self):
        result1 = DeepHash(Fraction(1, 2))
        result2 = DeepHash(0.5)
        assert result1[Fraction(1, 2)] != result2[0.5]

    def test_fraction_vs_float_hash_same_with_ignore_numeric_type(self):
        result1 = DeepHash(Fraction(1, 2), ignore_numeric_type_changes=True)
        result2 = DeepHash(0.5, ignore_numeric_type_changes=True)
        assert result1[Fraction(1, 2)] == result2[0.5]

    def test_fraction_hash_prep(self):
        result = DeepHashPrep(Fraction(1, 3))
        assert 'Fraction' in result[Fraction(1, 3)]

    def test_fraction_hash_prep_ignore_numeric_type(self):
        result = DeepHashPrep(Fraction(1, 2), ignore_numeric_type_changes=True)
        assert 'number' in result[Fraction(1, 2)]

    def test_fraction_hash_significant_digits(self):
        r1 = DeepHashPrep(Fraction(1, 3), significant_digits=2)
        r2 = DeepHashPrep(Fraction(334, 1000), significant_digits=2)
        assert r1[Fraction(1, 3)] == r2[Fraction(334, 1000)]


class TestFractionSerialization:
    """Tests for JSON serialization of Fraction values."""

    def test_fraction_to_json(self):
        t1 = Fraction(1, 3)
        t2 = Fraction(2, 3)
        ddiff = DeepDiff(t1, t2)
        json_str = ddiff.to_json()
        assert json_str is not None
        assert '"new_value"' in json_str

    def test_fraction_integer_value_serialization(self):
        """Fraction with denominator 1 should serialize as int."""
        result = json_dumps(Fraction(5, 1))
        assert result == '5'

    def test_fraction_float_value_serialization(self):
        """Fraction with denominator != 1 should serialize as float."""
        result = json_dumps(Fraction(1, 2))
        assert result == '0.5'

    def test_fraction_json_roundtrip(self):
        t1 = {"a": Fraction(1, 2), "b": [1, 2]}
        t2 = {"a": Fraction(3, 4), "b": [1, 3]}
        ddiff = DeepDiff(t1, t2)
        json_str = ddiff.to_json()
        loaded = json_loads(json_str)
        assert loaded is not None



#!/usr/bin/env python
import pytest
from deepdiff import DeepDiff


class TestMemoryView:
    """Test memoryview support in DeepDiff"""

    def test_memoryview_basic_comparison(self):
        """Test basic memoryview comparison without ignore_string_type_changes"""
        t1 = memoryview(b"hello")
        t2 = memoryview(b"world")
        
        diff = DeepDiff(t1, t2)
        assert 'values_changed' in diff
        assert diff['values_changed']['root']['old_value'] == t1
        assert diff['values_changed']['root']['new_value'] == t2
    
    def test_memoryview_with_bytes_type_change(self):
        """Test memoryview vs bytes comparison shows type change"""
        t1 = memoryview(b"hello")
        t2 = b"hello"
        
        diff = DeepDiff(t1, t2)
        assert 'type_changes' in diff
        assert diff['type_changes']['root']['old_type'] == memoryview
        assert diff['type_changes']['root']['new_type'] == bytes
        assert diff['type_changes']['root']['old_value'] == t1
        assert diff['type_changes']['root']['new_value'] == t2
    
    def test_memoryview_with_str_type_change(self):
        """Test memoryview vs str comparison shows type change"""
        t1 = memoryview(b"hello")
        t2 = "hello"
        
        diff = DeepDiff(t1, t2)
        assert 'type_changes' in diff
        assert diff['type_changes']['root']['old_type'] == memoryview
        assert diff['type_changes']['root']['new_type'] == str
        assert diff['type_changes']['root']['old_value'] == t1
        assert diff['type_changes']['root']['new_value'] == t2
    
    def test_memoryview_ignore_string_type_changes_with_bytes(self):
        """Test memoryview vs bytes with ignore_string_type_changes=True"""
        t1 = memoryview(b"hello")
        t2 = b"hello"
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_ignore_string_type_changes_with_str(self):
        """Test memoryview vs str with ignore_string_type_changes=True"""
        t1 = memoryview(b"hello")
        t2 = "hello"
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_different_content_with_ignore_string_type_changes(self):
        """Test memoryview with different content still shows value change"""
        t1 = memoryview(b"hello")
        t2 = "world"
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert 'values_changed' in diff
        # The values in the diff are the original objects, not converted strings
        assert diff['values_changed']['root']['old_value'] == t1
        assert diff['values_changed']['root']['new_value'] == t2
    
    def test_memoryview_in_dict_keys(self):
        """Test memoryview as dictionary keys"""
        t1 = {memoryview(b"key1"): "value1", memoryview(b"key2"): "value2"}
        t2 = {b"key1": "value1", "key2": "value2"}
        
        # Without ignore_string_type_changes, should show differences
        diff = DeepDiff(t1, t2)
        assert 'dictionary_item_removed' in diff or 'dictionary_item_added' in diff
        
        # With ignore_string_type_changes, should be equal
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_in_list(self):
        """Test memoryview in lists"""
        t1 = [memoryview(b"hello"), memoryview(b"world")]
        t2 = ["hello", b"world"]
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_in_nested_structure(self):
        """Test memoryview in nested structures"""
        t1 = {
            "data": {
                "items": [memoryview(b"item1"), memoryview(b"item2")],
                "metadata": {memoryview(b"key"): "value"}
            }
        }
        t2 = {
            "data": {
                "items": ["item1", b"item2"],
                "metadata": {"key": "value"}
            }
        }
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_with_non_ascii_bytes(self):
        """Test memoryview with non-ASCII bytes"""
        t1 = memoryview(b"\x80\x81\x82")
        t2 = b"\x80\x81\x82"
        
        diff = DeepDiff(t1, t2, ignore_string_type_changes=True)
        assert diff == {}
    
    def test_memoryview_text_diff(self):
        """Test that text diff works with memoryview"""
        t1 = {"data": memoryview(b"hello\nworld")}
        t2 = {"data": memoryview(b"hello\nearth")}
        
        diff = DeepDiff(t1, t2)
        assert 'values_changed' in diff
        assert "root['data']" in diff['values_changed']
        # Should contain diff output
        assert 'diff' in diff['values_changed']["root['data']"]
    
    def test_memoryview_with_ignore_type_in_groups(self):
        """Test memoryview with ignore_type_in_groups parameter"""
        from deepdiff.helper import strings
        
        t1 = memoryview(b"hello")
        t2 = "hello"
        
        # Using ignore_type_in_groups with strings tuple
        diff = DeepDiff(t1, t2, ignore_type_in_groups=[strings])
        assert diff == {}
    
    def test_memoryview_hash(self):
        """Test that DeepHash works with memoryview"""
        from deepdiff import DeepHash
        
        # Test basic hashing
        obj1 = memoryview(b"hello")
        hash1 = DeepHash(obj1)
        assert hash1[obj1]
        
        # Test with ignore_string_type_changes
        obj2 = "hello"
        hash2 = DeepHash(obj2, ignore_string_type_changes=True)
        hash1_ignore = DeepHash(obj1, ignore_string_type_changes=True)
        
        # When ignoring string type changes, memoryview and str of same content should hash the same
        assert hash1_ignore[obj1] == hash2[obj2]


import pytest
from deepdiff import DeepDiff

t1 = {
    "foo": {
        "bar": {
            "veg": "potato",
            "fruit": "apple"
        }
    },
    "ingredients": [
        {
            "lunch": [
                "bread",
                "cheese"
            ]
        },
        {
            "dinner": [
                "soup",
                "meat"
            ]
        }
    ]
}
t2 = {
    "foo": {
        "bar": {
            "veg": "potato",
            "fruit": "peach"
        }
    },
    "ingredients": [
        {
            "lunch": [
                "bread",
                "cheese"
            ]
        },
        {
            "dinner": [
                "soup",
                "meat"
            ]
        }
    ]
}


class TestDeepDiffIncludePaths:

    @staticmethod
    def deep_diff(dict1, dict2, include_paths):
        diff = DeepDiff(dict1, dict2, include_paths=include_paths)
        print(diff)
        return diff

    def test_include_paths_root_neg(self):
        expected = {'values_changed': {"root['foo']['bar']['fruit']": {'new_value': 'peach', 'old_value': 'apple'}}}
        actual = self.deep_diff(t1, t2, 'foo')
        assert expected == actual

    def test_include_paths_root_pos(self):
        expected = {}
        actual = self.deep_diff(t1, t2, 'ingredients')
        assert expected == actual

    def test_include_paths_nest00_neg(self):
        expected = {'values_changed': {"root['foo']['bar']['fruit']": {'new_value': 'peach', 'old_value': 'apple'}}}
        actual = self.deep_diff(t1, t2, "root['foo']['bar']")
        assert expected == actual

    def test_include_paths_nest01_neg(self):
        expected = {'values_changed': {"root['foo']['bar']['fruit']": {'new_value': 'peach', 'old_value': 'apple'}}}
        actual = self.deep_diff(t1, t2, "root['foo']['bar']['fruit']")
        assert expected == actual

    def test_include_paths_nest_pos(self):
        expected = {}
        actual = self.deep_diff(t1, t2, "root['foo']['bar']['veg']")
        assert expected == actual

    @pytest.mark.parametrize(
        "test_num, data",
        [
            (
                1,  # test_num
                {
                    "old": {
                        'name': 'Testname Old',
                        'desciption': 'Desc Old',
                        'sub_path': {
                            'name': 'Testname Subpath old',
                            'desciption': 'Desc Subpath old',
                        },
                    },
                    "new": {
                        'name': 'Testname New',
                        'desciption': 'Desc New',
                        'new_attribute': 'New Value',
                        'sub_path': {
                            'name': 'Testname Subpath old',
                            'desciption': 'Desc Subpath old',
                        },
                    },
                    "include_paths": "root['sub_path']",
                    "expected_result1": {'dictionary_item_added': ["root['new_attribute']"], 'values_changed': {"root['name']": {'new_value': 'Testname New', 'old_value': 'Testname Old'}, "root['desciption']": {'new_value': 'Desc New', 'old_value': 'Desc Old'}}},
                    "expected_result2": {},
                },
            ),
            (
                2,  # test_num
                {
                    "old": {
                        'name': 'Testname Old',
                        'desciption': 'Desc Old',
                        'sub_path': {
                            'name': 'Testname Subpath old',
                            'desciption': 'Desc Subpath old',
                        },
                    },
                    "new": {
                        'name': 'Testname New',
                        'desciption': 'Desc New',
                        'new_attribute': 'New Value',
                        'sub_path': {
                            'name': 'Testname Subpath New',
                            'desciption': 'Desc Subpath old',
                        },
                    },
                    "include_paths": "root['sub_path']",
                    "expected_result1": {'dictionary_item_added': ["root['new_attribute']"], 'values_changed': {"root['name']": {'new_value': 'Testname New', 'old_value': 'Testname Old'}, "root['desciption']": {'new_value': 'Desc New', 'old_value': 'Desc Old'}, "root['sub_path']['name']": {'new_value': 'Testname Subpath New', 'old_value': 'Testname Subpath old'}}},
                    "expected_result2": {"values_changed": {"root['sub_path']['name']": {"old_value": "Testname Subpath old", "new_value": "Testname Subpath New"}}},
                },
            ),
            (
                3,  # test_num
                {
                    "old": {
                        'name': 'Testname Old',
                        'desciption': 'Desc Old',
                        'sub_path': {
                            'name': 'Testname Subpath old',
                            'desciption': 'Desc Subpath old',
                            'old_attr': 'old attr value',
                        },
                    },
                    "new": {
                        'name': 'Testname New',
                        'desciption': 'Desc New',
                        'new_attribute': 'New Value',
                        'sub_path': {
                            'name': 'Testname Subpath old',
                            'desciption': 'Desc Subpath New',
                            'new_sub_path_attr': 'new sub path attr value',
                        },
                    },
                    "include_paths": "root['sub_path']['name']",
                    "expected_result1": {'dictionary_item_added': ["root['new_attribute']", "root['sub_path']['new_sub_path_attr']"], 'dictionary_item_removed': ["root['sub_path']['old_attr']"], 'values_changed': {"root['name']": {'new_value': 'Testname New', 'old_value': 'Testname Old'}, "root['desciption']": {'new_value': 'Desc New', 'old_value': 'Desc Old'}, "root['sub_path']['desciption']": {'new_value': 'Desc Subpath New', 'old_value': 'Desc Subpath old'}}},
                    "expected_result2": {},
                },
            ),
            (
                4,  # test_num
                {
                    "old": {
                        'name': 'Testname old',
                        'desciption': 'Desc old',
                        'new_attribute': 'old Value',
                        'sub_path': {
                            'name': 'Testname',
                            'removed_attr': 'revemod attr value',
                        },
                    },
                    "new": {
                        'name': 'Testname new',
                        'desciption': 'Desc new',
                        'new_attribute': 'new Value',
                        'sub_path': {
                            'added_attr': 'Added Attr Value',
                            'name': 'Testname',
                        },
                    },
                    "include_paths": "root['sub_path']['name']",
                    "expected_result1": {'dictionary_item_added': ["root['sub_path']['added_attr']"], 'dictionary_item_removed': ["root['sub_path']['removed_attr']"], 'values_changed': {"root['name']": {'new_value': 'Testname new', 'old_value': 'Testname old'}, "root['desciption']": {'new_value': 'Desc new', 'old_value': 'Desc old'}, "root['new_attribute']": {'new_value': 'new Value', 'old_value': 'old Value'}}},
                    "expected_result2": {},
                },
            ),
            (
                5,  # test_num
                {
                    "old": {
                        'name': 'Testname',
                        'removed_attr': 'revemod attr value',
                    },
                    "new": {
                        'added_attr': 'Added Attr Value',
                        'name': 'Testname',
                    },
                    "include_paths": "root['name']",
                    "expected_result1": {'dictionary_item_added': ["root['added_attr']"], 'dictionary_item_removed': ["root['removed_attr']"]},
                    "expected_result2": {},
                },
            ),
            (
                6,  # test_num
                {
                    "old": {
                        'name': 'Testname',
                        'removed_attr': 'revemod attr value',
                        'removed_attr_2': 'revemod attr value',
                    },
                    "new": {
                        'added_attr': 'Added Attr Value',
                        'name': 'Testname',
                    },
                    "include_paths": "root['name']",
                    "expected_result1": {'values_changed': {'root': {'new_value': {'added_attr': 'Added Attr Value', 'name': 'Testname'}, 'old_value': {'name': 'Testname', 'removed_attr': 'revemod attr value', 'removed_attr_2': 'revemod attr value'}}}},
                    "expected_result2": {},
                },
            ),
            (
                7,  # test_num
                {
                    "old": {
                        'name': 'Testname old',
                        'desciption': 'Desc old',
                        'new_attribute': 'old Value',
                        'sub_path': {
                            'name': 'Testname',
                            'removed_attr': 'revemod attr value',
                            'removed_attr_2': 'blu',
                        },
                    },
                    "new": {
                        'name': 'Testname new',
                        'desciption': 'Desc new',
                        'new_attribute': 'new Value',
                        'sub_path': {
                            'added_attr': 'Added Attr Value',
                            'name': 'Testname',
                        },
                    },
                    "include_paths": "root['sub_path']['name']",
                    "expected_result1": {'values_changed': {"root['name']": {'new_value': 'Testname new', 'old_value': 'Testname old'}, "root['desciption']": {'new_value': 'Desc new', 'old_value': 'Desc old'}, "root['new_attribute']": {'new_value': 'new Value', 'old_value': 'old Value'}, "root['sub_path']": {'new_value': {'added_attr': 'Added Attr Value', 'name': 'Testname'}, 'old_value': {'name': 'Testname', 'removed_attr': 'revemod attr value', 'removed_attr_2': 'blu'}}}},
                    "expected_result2": {},
                },
            ),
            (
                8,  # test_num
                {
                    "old": [{
                        'name': 'Testname old',
                        'desciption': 'Desc old',
                        'new_attribute': 'old Value',
                        'sub_path': {
                            'name': 'Testname',
                            'removed_attr': 'revemod attr value',
                            'removed_attr_2': 'blu',
                        },
                    }],
                    "new": [{
                        'name': 'Testname new',
                        'desciption': 'Desc new',
                        'new_attribute': 'new Value',
                        'sub_path': {
                            'added_attr': 'Added Attr Value',
                            'name': 'New Testname',
                        },
                    }],
                    "include_paths": "root[0]['sub_path']['name']",
                    "expected_result1": {'values_changed': {"root[0]['name']": {'new_value': 'Testname new', 'old_value': 'Testname old'}, "root[0]['desciption']": {'new_value': 'Desc new', 'old_value': 'Desc old'}, "root[0]['new_attribute']": {'new_value': 'new Value', 'old_value': 'old Value'}, "root[0]['sub_path']": {'new_value': {'added_attr': 'Added Attr Value', 'name': 'New Testname'}, 'old_value': {'name': 'Testname', 'removed_attr': 'revemod attr value', 'removed_attr_2': 'blu'}}}},
                    "expected_result2": {'values_changed': {"root[0]['sub_path']['name']": {'new_value': 'New Testname', 'old_value': 'Testname'}}},
                },
            ),
        ]
    )
    def test_diff_include_paths_root(self, test_num, data):
        diff1 = DeepDiff(data["old"], data["new"])
        diff2 = DeepDiff(data["old"], data["new"], include_paths=data["include_paths"])
        assert data['expected_result1'] == diff1, f"test_diff_include_paths_root test_num #{test_num} failed."
        assert data['expected_result2'] == diff2, f"test_diff_include_paths_root test_num #{test_num} failed."



#!/usr/bin/env python
from ipaddress import IPv4Address
import os
import json
import sys
import pytest
import datetime
import numpy as np
import hashlib
import base64
from typing import NamedTuple, Optional, Union, List, Dict
from pickle import UnpicklingError
from decimal import Decimal
from collections import Counter
from pydantic import BaseModel, IPvAnyAddress
from deepdiff import DeepDiff
from deepdiff.helper import pypy3, py_current_version, np_ndarray, Opcode, SetOrdered
from deepdiff.serialization import (
    pickle_load, pickle_dump, ForbiddenModule, ModuleNotFoundError,
    MODULE_NOT_FOUND_MSG, FORBIDDEN_MODULE_MSG, pretty_print_diff,
    load_path_content, UnsupportedFormatErr, json_dumps, json_loads)
from conftest import FIXTURES_DIR
from tests import PicklableClass

import logging
logging.disable(logging.CRITICAL)

t1 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 2, 3]}}
t2 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": "world\n\n\nEnd"}}


class SampleSchema(BaseModel):
    works: bool = False
    ips: List[IPvAnyAddress]


class SomeStats(NamedTuple):
    counter: Optional[Counter]
    context_aware_counter: Optional[Counter] = None
    min_int: Optional[int] = 0
    max_int: Optional[int] = 0


field_stats1 = SomeStats(
    counter=Counter(["a", "a", "b"]),
    max_int=10
)


class TestSerialization:
    """Tests for Serializations."""

    def test_serialization_text(self):
        ddiff = DeepDiff(t1, t2)
        assert "builtins.list" in ddiff.to_json_pickle()
        jsoned = ddiff.to_json()
        assert "world" in jsoned

    def test_serialization_text_force_builtin_json(self):
        ddiff = DeepDiff(t1, t2)
        with pytest.raises(TypeError) as excinfo:
            jsoned = ddiff.to_json(sort_keys=True)
        assert str(excinfo.value).startswith("orjson does not accept the sort_keys parameter.")
        jsoned = ddiff.to_json(sort_keys=True, force_use_builtin_json=True)
        assert "world" in jsoned

    def test_deserialization(self):
        ddiff = DeepDiff(t1, t2)
        jsoned = ddiff.to_json_pickle()
        ddiff2 = DeepDiff.from_json_pickle(jsoned)
        assert ddiff == ddiff2

    def test_serialization_tree(self):
        ddiff = DeepDiff(t1, t2, view='tree')
        pickle_jsoned = ddiff.to_json_pickle()
        assert "world" in pickle_jsoned
        jsoned = ddiff.to_json()
        assert "world" in jsoned

    def test_deserialization_tree(self):
        ddiff = DeepDiff(t1, t2, view='tree')
        jsoned = ddiff.to_json_pickle()
        ddiff2 = DeepDiff.from_json_pickle(jsoned)
        assert 'type_changes' in ddiff2

    def test_serialize_custom_objects_throws_error(self):
        class A:
            pass

        class B:
            pass

        t1 = A()
        t2 = B()
        ddiff = DeepDiff(t1, t2)
        assert r'{"type_changes":{"root":{"old_type":"A","new_type":"B","old_value":{},"new_value":{}}}}' == ddiff.to_json()

    def test_serialize_custom_objects_with_default_mapping(self):
        class A:
            pass

        class B:
            pass

        t1 = A()
        t2 = B()
        ddiff = DeepDiff(t1, t2)
        default_mapping = {A: lambda x: 'obj A', B: lambda x: 'obj B'}
        result = ddiff.to_json(default_mapping=default_mapping)
        expected_result = {"type_changes": {"root": {"old_type": "A", "new_type": "B", "old_value": "obj A", "new_value": "obj B"}}}
        assert expected_result == json.loads(result)

    # These lines are long but make it easier to notice the difference:
    @pytest.mark.parametrize('verbose_level, expected', [
        (0, {"type_changes": {"root[0]": {"old_type": str, "new_type": int}}, "dictionary_item_added": ["root[1][5]"], "dictionary_item_removed": ["root[1][3]"], "iterable_item_added": {"root[2]": "d"}}),
        (1, {"type_changes": {"root[0]": {"old_type": str, "new_type": int, "old_value": "a", "new_value": 10}}, "dictionary_item_added": ["root[1][5]"], "dictionary_item_removed": ["root[1][3]"], "values_changed": {"root[1][1]": {"new_value": 2, "old_value": 1}}, "iterable_item_added": {"root[2]": "d"}}),
        (2, {"type_changes": {"root[0]": {"old_type": str, "new_type": int, "old_value": "a", "new_value": 10}}, "dictionary_item_added": {"root[1][5]": 6}, "dictionary_item_removed": {"root[1][3]": 4}, "values_changed": {"root[1][1]": {"new_value": 2, "old_value": 1}}, "iterable_item_added": {"root[2]": "d"}}),
    ])
    def test_to_dict_at_different_verbose_level(self, verbose_level, expected):
        t1 = ['a', {1: 1, 3: 4}, ]
        t2 = [10, {1: 2, 5: 6}, 'd']

        ddiff = DeepDiff(t1, t2, verbose_level=verbose_level)
        assert expected == ddiff.to_dict()

    def test_to_dict_tree_view_defaults_to_verbose_level_2(self):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        ddiff = DeepDiff(t1, t2, view='tree')
        result = ddiff.to_dict()
        # tree view defaults to verbose_level=2: added/removed are dicts, not sets
        expected = {
            "type_changes": {"root[0]": {"old_type": str, "new_type": int, "old_value": "a", "new_value": 10}},
            "dictionary_item_added": {"root[1][5]": 6},
            "dictionary_item_removed": {"root[1][3]": 4},
            "values_changed": {"root[1][1]": {"new_value": 2, "old_value": 1}},
            "iterable_item_added": {"root[2]": "d"},
        }
        assert expected == result

    @pytest.mark.parametrize('verbose_level, expected', [
        (0, {"type_changes": {"root[0]": {"old_type": str, "new_type": int}}, "dictionary_item_added": ["root[1][5]"], "dictionary_item_removed": ["root[1][3]"], "iterable_item_added": {"root[2]": "d"}}),
        (1, {"type_changes": {"root[0]": {"old_type": str, "new_type": int, "old_value": "a", "new_value": 10}}, "dictionary_item_added": ["root[1][5]"], "dictionary_item_removed": ["root[1][3]"], "values_changed": {"root[1][1]": {"new_value": 2, "old_value": 1}}, "iterable_item_added": {"root[2]": "d"}}),
    ])
    def test_to_dict_tree_view_with_verbose_level_override(self, verbose_level, expected):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        ddiff = DeepDiff(t1, t2, view='tree')
        result = ddiff.to_dict(verbose_level=verbose_level)
        assert expected == result

    def test_to_dict_text_view_preserves_original_verbose_level(self):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        # verbose_level=2 at init, text view
        ddiff = DeepDiff(t1, t2, verbose_level=2)
        result = ddiff.to_dict()
        # Should preserve verbose_level=2 from init
        assert isinstance(result.get("dictionary_item_added"), dict)
        assert result["dictionary_item_added"] == {"root[1][5]": 6}

    def test_to_dict_text_view_with_verbose_level_override(self):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        # verbose_level=2 at init, but override to 1 in to_dict
        ddiff = DeepDiff(t1, t2, verbose_level=2)
        result = ddiff.to_dict(verbose_level=1)
        # verbose_level=1: dictionary_item_added is not a dict (it's a SetOrdered)
        assert not isinstance(result.get("dictionary_item_added"), dict)

    def test_to_dict_invalid_verbose_level(self):
        t1 = [1]
        t2 = [2]
        ddiff = DeepDiff(t1, t2)
        with pytest.raises(ValueError):
            ddiff.to_dict(verbose_level=3)

    def test_to_json_with_verbose_level(self):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        ddiff = DeepDiff(t1, t2, view='tree')
        result = json.loads(ddiff.to_json())
        # tree view defaults to verbose_level=2 in to_json too
        assert "root[1][5]" in result.get("dictionary_item_added", {})

    def test_to_json_with_verbose_level_override(self):
        t1 = ['a', {1: 1, 3: 4}]
        t2 = [10, {1: 2, 5: 6}, 'd']
        ddiff = DeepDiff(t1, t2, view='tree')
        result = json.loads(ddiff.to_json(verbose_level=1))
        # verbose_level=1: dictionary_item_added is a list
        assert isinstance(result.get("dictionary_item_added"), list)

    def test_serialize_pydantic_model(self):
        obj = SampleSchema(
            works=True,
            ips=["128.0.0.1"]
        )
        serialized = json_dumps(obj)
        obj_again = json_loads(serialized)
        assert {'works': True, 'ips': ['128.0.0.1']} == obj_again
        assert {'works': True, 'ips': [IPv4Address('128.0.0.1')]} == obj.model_dump()


@pytest.mark.skipif(pypy3, reason='clevercsv is not supported in pypy3')
class TestLoadContet:

    @pytest.mark.parametrize('path1, validate', [
        ('t1.json', lambda x: x[0]['key1'] == 'value1'),
        ('t1.yaml', lambda x: x[0][0] == 'name'),
        ('t1.toml', lambda x: x['servers']['alpha']['ip'] == '10.0.0.1'),
        ('t1.csv', lambda x: x[0]['last_name'] == 'Nobody'),
        ('t1.pickle', lambda x: x[1] == 1),
    ])
    def test_load_path_content(self, path1, validate):
        path = os.path.join(FIXTURES_DIR, path1)
        result = load_path_content(path)
        assert validate(result)

    def test_load_path_content_when_unsupported_format(self):
        path = os.path.join(FIXTURES_DIR, 't1.unsupported')
        with pytest.raises(UnsupportedFormatErr):
            load_path_content(path)


class TestPicklingSecurity:

    @pytest.mark.skipif(sys.platform == "win32", reason="Resource module is Unix-only")
    def test_restricted_unpickler_memory_exhaustion_cve(self):
        """CVE-2026-33155: Prevent DoS via massive allocation through REDUCE opcode.

        The payload calls bytes(10_000_000_000) which is allowed by find_class
        but would allocate ~9.3GB of memory. The fix should reject this before
        the allocation happens.
        """
        import resource

        # 1. Cap memory to 500MB to prevent system freezes during the test
        soft, hard = resource.getrlimit(resource.RLIMIT_AS)
        maxsize_bytes = 500 * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (maxsize_bytes, hard))

        try:
            # 2. Malicious payload: attempts to allocate ~9.3GB via bytes(10000000000)
            # This uses allowed builtins but passes a massive integer via REDUCE
            payload = (
                b"(dp0\n"
                b"S'_'\n"
                b"cbuiltins\nbytes\n"
                b"(I10000000000\n"
                b"tR"
                b"s."
            )

            # 3. After the patch, deepdiff should catch the size violation
            # and raise UnpicklingError before attempting allocation.
            with pytest.raises((ValueError, UnpicklingError)):
                pickle_load(payload)
        finally:
            # Restore original memory limit so other tests are not affected
            resource.setrlimit(resource.RLIMIT_AS, (soft, hard))

    def test_restricted_unpickler_allows_small_bytes(self):
        """Ensure legitimate small bytes objects can still be deserialized."""
        # Payload: {'_': bytes(100)} — well within the 128MB limit
        payload = (
            b"(dp0\n"
            b"S'_'\n"
            b"cbuiltins\nbytes\n"
            b"(I100\n"
            b"tR"
            b"s."
        )
        result = pickle_load(payload)
        assert result == {'_': bytes(100)}


class TestPickling:

    def test_serialize(self):
        obj = [1, 2, 3, None, {10: 11E2}, frozenset(['a', 'c']), SetOrdered([2, 1]),
               datetime.datetime(2022, 4, 10, 0, 40, 41, 357857), datetime.time(11), Decimal('11.2'), 123.11]
        serialized = pickle_dump(obj)
        loaded = pickle_load(serialized)
        assert obj == loaded

    @pytest.mark.skipif(pypy3, reason='short pickle not supported in pypy3')
    def test_pickle_that_is_string(self):
        serialized_str = 'DeepDiff Delta Payload v0-0-1\nBlah'
        with pytest.raises(UnpicklingError):
            pickle_load(serialized_str)

    def test_custom_object_deserialization_fails_without_explicit_permission(self):
        obj = PicklableClass(10)
        module_dot_name = 'tests.{}'.format(PicklableClass.__name__)

        serialized = pickle_dump(obj)

        expected_msg = FORBIDDEN_MODULE_MSG.format(module_dot_name)
        with pytest.raises(ForbiddenModule) as excinfo:
            pickle_load(serialized)
        assert expected_msg == str(excinfo.value)

        # Explicitly allowing the module to be loaded
        loaded = pickle_load(serialized, safe_to_import={module_dot_name})
        assert obj == loaded

        # Explicitly allowing the module to be loaded. It can take a list instead of a set.
        loaded2 = pickle_load(serialized, safe_to_import=[module_dot_name])
        assert obj == loaded2

    def test_unpickling_object_that_is_not_imported_raises_error(self):

        def get_the_pickle():
            import wave
            obj = wave.Error
            return pickle_dump(obj)

        serialized = get_the_pickle()
        # Making sure that the module is unloaded.
        del sys.modules['wave']
        module_dot_name = 'wave.Error'

        expected_msg = MODULE_NOT_FOUND_MSG.format(module_dot_name)
        with pytest.raises(ModuleNotFoundError) as excinfo:
            pickle_load(serialized, safe_to_import=module_dot_name)
        assert expected_msg == str(excinfo.value)


    def test_seriaize_property(self):

        class Sample:
            @property
            def something(self):
                return 10

        sample = Sample()
        serialized = json_dumps(sample)
        assert '{"something":10}' == serialized


class TestDeepDiffPretty:
    """Tests for pretty() method of DeepDiff"""

    class TestingClass:
        one = 1

    testing_class = TestingClass

    @pytest.mark.parametrize('t1, t2, item_path, old_type, new_type, old_val_displayed, new_val_displayed',
                             [
                                 [{2: 2, 4: 4}, {2: 'b', 4: 4}, 'root[2]', 'int', 'str', '2', '"b"'],
                                 [[1, 2, 3], [1, '2', 3], 'root[1]', 'int', 'str', '2', '"2"'],
                                 [[1, 2, 3], {1, 2, 3}, 'root', 'list', 'set', '[1, 2, 3]', '{1, 2, 3}']
                             ])
    def test_pretty_print_diff_type_changes(self, t1, t2, item_path, old_type, new_type, old_val_displayed,
                                            new_val_displayed):
        ddiff = DeepDiff(t1, t2, view='tree')
        result = pretty_print_diff(ddiff.tree['type_changes'][0])
        assert result == 'Type of {} changed from {} to {} and value changed from {} to {}.'.format(item_path, old_type, new_type, old_val_displayed, new_val_displayed)

    @pytest.mark.parametrize('t1, t2, item_path, verbose_level',
                             [
                                 [{2: 2, 4: 4}, {2: 2, 4: 4, 5: 5}, 'root[5]', 1],
                                 [{2: 2, 4: 4}, {2: 2, 4: 4, 5: 5}, 'root[5] (5)', 2],
                                 [{"foo": "bar", "foo1": "bar1"}, {"foo": "bar", "foo1": "bar1", "foo2": "bar2"},
                                  'root[\'foo2\']', 0],
                                 [{"foo": "bar", "foo1": "bar1"}, {"foo": "bar", "foo1": "bar1", "foo2": "bar2"},
                                  'root[\'foo2\'] ("bar2")', 2]
                             ])
    def test_pretty_print_diff_dictionary_item_added(self, t1, t2, item_path, verbose_level):
        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['dictionary_item_added'][0])
        assert result == 'Item {} added to dictionary.'.format(item_path)

    @pytest.mark.parametrize('t1, t2, item_path, verbose_level',
                             [
                                 [{2: 2, 4: 4}, {2: 2}, 'root[4]', 0],
                                 [{2: 2, 4: 4}, {2: 2}, 'root[4] (4)', 2],
                                 [{"foo": "bar", "foo1": "bar1"}, {"foo": "bar"},
                                  'root[\'foo1\']', 1],
                                 [{"foo": "bar", "foo1": "bar1"}, {"foo": "bar"},
                                  'root[\'foo1\'] ("bar1")', 2],
                             ])
    def test_pretty_print_diff_dictionary_item_removed(self, t1, t2, item_path, verbose_level):
        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['dictionary_item_removed'][0])
        assert result == 'Item {} removed from dictionary.'.format(item_path)

    @pytest.mark.parametrize('t1, t2, item_path, old_val_displayed, new_val_displayed',
                             [
                                 [{2: 2, 4: 4}, {2: 3, 4: 4}, 'root[2]', '2', '3'],
                                 [['a', 'b', 'c'], ['a', 'b', 'd'], 'root[2]', '"c"', '"d"']
                             ])
    def test_pretty_print_diff_values_changed(self, t1, t2, item_path, old_val_displayed, new_val_displayed):
        ddiff = DeepDiff(t1, t2, view='tree')
        result = pretty_print_diff(ddiff.tree['values_changed'][0])
        assert result == 'Value of {} changed from {} to {}.'.format(item_path, old_val_displayed, new_val_displayed)

    @pytest.mark.parametrize('t1, t2, item_path, verbose_level',
                             [
                                 [[1, 2, 3], [1, 2, 3, 4], 'root[3]', 1],
                                 [[1, 2, 3], [1, 2, 3, 4], 'root[3] (4)', 2],
                                 [["foo", "bar"], ["foo", "bar", "barbar"], 'root[2]', 0],
                                 [["foo", "bar"], ["foo", "bar", "barbar"], 'root[2] ("barbar")', 2]
                             ])
    def test_pretty_print_diff_iterable_item_added(self, t1, t2, item_path, verbose_level):
        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['iterable_item_added'][0])
        assert result == 'Item {} added to iterable.'.format(item_path)

    @pytest.mark.parametrize('t1, t2, item_path, verbose_level',
                             [
                                 [[1, 2, 3], [1, 2], 'root[2]', 0],
                                 [[1, 2, 3], [1, 2], 'root[2] (3)', 2],
                                 [["foo", "bar", "barbar"], ["foo", "bar"], 'root[2]', 1],
                                 [["foo", "bar", "barbar"], ["foo", "bar"], 'root[2] ("barbar")', 2]
                             ])
    def test_pretty_print_diff_iterable_item_removed(self, t1, t2, item_path, verbose_level):
        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['iterable_item_removed'][0])
        assert result == 'Item {} removed from iterable.'.format(item_path)

    @pytest.mark.parametrize("verbose_level", range(3))
    def test_pretty_print_diff_attribute_added(self, verbose_level):
        t1 = self.testing_class()
        t2 = self.testing_class()
        t2.two = 2

        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['attribute_added'][0])
        assert result == 'Attribute root.two (2) added.' if verbose_level == 2 else 'Attribute root.two added.'

    @pytest.mark.parametrize("verbose_level", range(3))
    def test_pretty_print_diff_attribute_removed(self, verbose_level):
        t1 = self.testing_class()
        t1.two = 2
        t2 = self.testing_class()

        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = pretty_print_diff(ddiff.tree['attribute_removed'][0])

        assert result == 'Attribute root.two (2) removed.' if verbose_level == 2 else 'Attribute root.two removed.'

    @pytest.mark.parametrize('t1, t2, item_path',
                             [
                                 [{1, 2}, {1, 2, 3}, 'root[3]'],
                             ])
    def test_pretty_print_diff_set_item_added(self, t1, t2, item_path):
        ddiff = DeepDiff(t1, t2, view='tree')
        result = pretty_print_diff(ddiff.tree['set_item_added'][0])
        assert result == 'Item {} added to set.'.format(item_path)

    @pytest.mark.parametrize('t1, t2, item_path',
                             [
                                 [{1, 2, 3}, {1, 2}, 'root[3]'],
                             ])
    def test_pretty_print_diff_set_item_removed(self, t1, t2, item_path):
        ddiff = DeepDiff(t1, t2, view='tree')
        result = pretty_print_diff(ddiff.tree['set_item_removed'][0])
        assert result == 'Item {} removed from set.'.format(item_path)

    @pytest.mark.parametrize('t1, t2, item_path',
                             [
                                 [[1, 2, 3, 2], [1, 2, 3], 'root[1]'],
                             ])
    def test_pretty_print_diff_repetition_change(self, t1, t2, item_path):
        ddiff = DeepDiff(t1, t2, view='tree', ignore_order=True, report_repetition=True)
        result = pretty_print_diff(ddiff.tree['repetition_change'][0])
        assert result == 'Repetition change for item {}.'.format(item_path)

    @pytest.mark.parametrize("expected, verbose_level",
                             (
                                 ('Item root[5] added to dictionary.'
                                  '\nItem root[3] removed from dictionary.'
                                  '\nType of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\nValue of root[4] changed from 4 to 5.', 0),
                                 ('Item root[5] (5) added to dictionary.'
                                  '\nItem root[3] (3) removed from dictionary.'
                                  '\nType of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\nValue of root[4] changed from 4 to 5.', 2),
                             ), ids=("verbose=0", "verbose=2")
                             )
    def test_pretty_form_method(self, expected, verbose_level):
        t1 = {2: 2, 3: 3, 4: 4}
        t2 = {2: 'b', 4: 5, 5: 5}
        ddiff = DeepDiff(t1, t2, view='tree', verbose_level=verbose_level)
        result = ddiff.pretty()
        assert result == expected

    @pytest.mark.parametrize("expected, verbose_level",
                             (
                                 ('\t\tItem root[5] added to dictionary.'
                                  '\n\t\tItem root[3] removed from dictionary.'
                                  '\n\t\tType of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\n\t\tValue of root[4] changed from 4 to 5.', 0),
                                 ('\t\tItem root[5] (5) added to dictionary.'
                                  '\n\t\tItem root[3] (3) removed from dictionary.'
                                  '\n\t\tType of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\n\t\tValue of root[4] changed from 4 to 5.', 2),
                             ), ids=("verbose=0", "verbose=2")
                             )
    def test_pretty_form_method_prefixed_simple(self, expected, verbose_level):
        t1 = {2: 2, 3: 3, 4: 4}
        t2 = {2: 'b', 4: 5, 5: 5}
        ddiff = DeepDiff(t1, t2, verbose_level=verbose_level)
        result = ddiff.pretty(prefix="\t\t")
        assert result == expected

    @pytest.mark.parametrize("expected, verbose_level",
                             (
                                 ('Diff #1: Item root[5] added to dictionary.'
                                  '\nDiff #2: Item root[3] removed from dictionary.'
                                  '\nDiff #3: Type of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\nDiff #4: Value of root[4] changed from 4 to 5.', 0),
                                 ('Diff #1: Item root[5] (5) added to dictionary.'
                                  '\nDiff #2: Item root[3] (3) removed from dictionary.'
                                  '\nDiff #3: Type of root[2] changed from int to str and value changed from 2 to "b".'
                                  '\nDiff #4: Value of root[4] changed from 4 to 5.', 2),
                             ), ids=("verbose=0", "verbose=2")
                             )
    def test_pretty_form_method_prefixed_callback(self, expected, verbose_level):
        def prefix_callback(**kwargs):
            """Helper function using a hidden variable on the diff that tracks which count prints next"""
            kwargs['diff']._diff_count = 1 + getattr(kwargs['diff'], '_diff_count', 0)
            return f"Diff #{kwargs['diff']._diff_count}: "

        t1 = {2: 2, 3: 3, 4: 4}
        t2 = {2: 'b', 4: 5, 5: 5}
        ddiff = DeepDiff(t1, t2, verbose_level=verbose_level)
        result = ddiff.pretty(prefix=prefix_callback)
        assert result == expected

    def sig_to_bytes(inp: Dict[str, Union[str, bytes]]):
        inp['signature'] = inp['signature'].encode('utf-8')
        return inp

    @pytest.mark.parametrize('test_num, value, func_to_convert_back', [
        (1, {'10': None}, None),
        (2, {"type_changes": {"root": {"old_type": None, "new_type": list, "new_value": ["你好", 2, 3, 5]}}}, None),
        (3, {'10': Decimal(2017)}, None),
        (4, Decimal(2017.1), None),
        (5, {1, 2, 10}, set),
        (6, datetime.datetime(2023, 10, 11), datetime.datetime.fromisoformat),
        (7, datetime.datetime.utcnow(), datetime.datetime.fromisoformat),
        (8, field_stats1, lambda x: SomeStats(**x)),
        (9, np.array([[ 101, 3533, 1998, 4532, 2024, 3415, 1012,  102]]), np.array),
        (10, memoryview(b"hello"), lambda x: memoryview(x.encode('utf-8'))),
        (11, {'file_type': 'xlsx', 'signature': b'52bd9907785'}, sig_to_bytes)
    ])
    def test_json_dumps_and_loads(self, test_num, value, func_to_convert_back):
        serialized = json_dumps(value)
        back = json_loads(serialized)
        if func_to_convert_back:
            back = func_to_convert_back(back)
        if isinstance(back, np_ndarray):
            assert np.array_equal(value, back), f"test_json_dumps_and_loads test #{test_num} failed"
        else:
            assert value == back, f"test_json_dumps_and_loads test #{test_num} failed"

    def test_namedtuple_seriazliation(self):
        op_code = Opcode(tag="replace", t1_from_index=0, t1_to_index=1, t2_from_index=10, t2_to_index=20)
        serialized = json_dumps(op_code)
        expected = '{"tag":"replace","t1_from_index":0,"t1_to_index":1,"t2_from_index":10,"t2_to_index":20,"old_values":null,"new_values":null}'
        assert serialized == expected

    def test_reversed_list(self):
        items = reversed([1, 2, 3])

        serialized = json_dumps(items)
        serialized2 = json_dumps(items)

        assert '[3,2,1]' == serialized
        assert '[3,2,1]' == serialized2, "We should have copied the original list. If this returns empty, it means we exhausted the original list."

    def test_dict_keys(self):
        dic = {"foo": "bar", "apple": "too sweet"}
        serialized = json_dumps(dic.keys())
        assert '["foo","apple"]' == serialized

    def test_non_utf8_bytes_serialization(self):
        """Test that non-UTF-8 bytes are properly base64 encoded"""
        # Create binary data that cannot be decoded as UTF-8
        binary_data = b'\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f'
        
        # Verify it's not UTF-8 decodable
        with pytest.raises(UnicodeDecodeError):
            binary_data.decode('utf-8')
        
        # Test serialization
        test_data = {"binary": binary_data}
        serialized = json_dumps(test_data)
        
        # Should contain base64 encoded data
        expected_b64 = base64.b64encode(binary_data).decode('ascii')
        assert expected_b64 in serialized
        
        # Should be deserializable
        deserialized = json_loads(serialized)
        assert deserialized == {"binary": expected_b64}

    def test_hash_bytes_serialization(self):
        """Test serialization of hash-like binary data (blake3, sha256, etc.)"""
        # Generate various hash-like byte sequences
        test_cases = [
            hashlib.md5(b"test").digest(),
            hashlib.sha1(b"test").digest(),
            hashlib.sha256(b"test").digest(),
            hashlib.sha512(b"test").digest()[:16],  # Truncated
            b'\xff\xfe\xfd\xfc' * 8,  # Artificial binary pattern
        ]
        
        for i, hash_bytes in enumerate(test_cases):
            test_data = {"hash": hash_bytes}
            
            # Should not raise UnicodeDecodeError
            serialized = json_dumps(test_data)
            assert serialized  # Should produce valid JSON
            
            # Should contain base64 if not UTF-8 decodable, or string if UTF-8 decodable
            try:
                utf8_decoded = hash_bytes.decode('utf-8')
                # If UTF-8 decodable, should be in JSON as string
                assert utf8_decoded in serialized
            except UnicodeDecodeError:
                # If not UTF-8 decodable, should be base64 encoded
                expected_b64 = base64.b64encode(hash_bytes).decode('ascii')
                assert expected_b64 in serialized

    def test_mixed_utf8_and_binary_bytes(self):
        """Test data structure with both UTF-8 decodable and binary bytes"""
        test_data = {
            "utf8_text": b"hello world",  # UTF-8 decodable
            "binary_hash": hashlib.sha256(b"secret").digest(),  # Binary
            "empty_bytes": b"",  # Edge case
            "utf8_unicode": "café".encode('utf-8'),  # UTF-8 with unicode
            "non_utf8_byte": b"\xff\xfe\xfd",  # Non-UTF-8 bytes
        }
        
        # Should serialize without errors
        serialized = json_dumps(test_data)
        deserialized = json_loads(serialized)
        
        # UTF-8 decodable bytes should remain as strings
        assert "hello world" in serialized
        assert deserialized["utf8_text"] == "hello world"
        
        # Unicode UTF-8 should work
        assert deserialized["utf8_unicode"] == "café"
        
        # Binary data should be base64 encoded
        expected_hash_b64 = base64.b64encode(test_data["binary_hash"]).decode('ascii')
        assert expected_hash_b64 in serialized
        assert deserialized["binary_hash"] == expected_hash_b64
        
        # Empty bytes should be empty string
        assert deserialized["empty_bytes"] == ""
        
        # Non-UTF-8 bytes should be base64 encoded
        expected_non_utf8_b64 = base64.b64encode(test_data["non_utf8_byte"]).decode('ascii')
        assert expected_non_utf8_b64 in serialized
        assert deserialized["non_utf8_byte"] == expected_non_utf8_b64

    def test_json_dumps_large_int_exceeding_64bit(self):
        """Test that integers exceeding 64-bit range are serialized as strings.
        orjson cannot handle integers larger than 64-bit (9223372036854775807),
        so json_dumps should convert them to strings."""
        large_int = 59579472846392086780
        assert large_int > 9223372036854775807  # larger than max int64

        data = {'max_int': large_int, 'min_int': 1336138}
        serialized = json_dumps(data)
        back = json_loads(serialized)
        assert back['max_int'] == str(large_int)
        assert back['min_int'] == 1336138

    def test_json_dumps_large_int_in_nested_structure(self):
        """Test large integers in nested data structures are converted to strings."""
        large_int = 59579472846392086780
        data = {
            'stats': {
                'counter': {'HasInt': 499},
                'max_int': large_int,
                'min_int': 1336138,
            },
            'name': 'policy_number',
        }
        serialized = json_dumps(data)
        back = json_loads(serialized)
        assert back['stats']['max_int'] == str(large_int)

    def test_json_dumps_large_negative_int(self):
        """Test that large negative integers are also converted to strings."""
        large_neg_int = -59579472846392086780
        data = {'value': large_neg_int}
        serialized = json_dumps(data)
        back = json_loads(serialized)
        assert back['value'] == str(large_neg_int)

    def test_json_dumps_namedtuple_with_large_int(self):
        """Test that a NamedTuple containing an oversized int is properly serialized.
        _convert_oversized_ints must reconstruct the NamedTuple via _fields, not
        pass a flat list to its constructor (which fails for NamedTuples with
        required keyword arguments)."""
        large_int = 59579472846392086780
        stats = SomeStats(
            counter=Counter(["a", "b"]),
            context_aware_counter=Counter(),
            min_int=0,
            max_int=large_int,
        )
        data = {'stats': stats}
        serialized = json_dumps(data)
        back = json_loads(serialized)
        assert back['stats']['max_int'] == str(large_int)
        assert back['stats']['min_int'] == 0

    def test_json_dumps_dict_of_namedtuples_with_large_int(self):
        """Test a dict of NamedTuples where one contains an oversized int.
        This mirrors the real-world pattern of field stats keyed by column name."""
        large_int = 59579472846392086780
        stats_map = {
            'normal_field': SomeStats(counter=Counter(["x"]), max_int=10),
            'big_field': SomeStats(counter=Counter(["y"]), max_int=large_int),
        }
        serialized = json_dumps(stats_map)
        back = json_loads(serialized)
        assert back['normal_field']['max_int'] == 10
        assert back['big_field']['max_int'] == str(large_int)

    def test_json_dumps_namedtuple_with_large_int_in_list(self):
        """Test a list of NamedTuples where one has an oversized int."""
        large_int = 59579472846392086780
        data = [
            SomeStats(counter=Counter(), max_int=5),
            SomeStats(counter=Counter(), max_int=large_int),
        ]
        serialized = json_dumps(data)
        back = json_loads(serialized)
        assert back[0]['max_int'] == 5
        assert back[1]['max_int'] == str(large_int)

    def test_bytes_in_deepdiff_serialization(self):
        """Test that bytes work correctly in DeepDiff JSON serialization"""
        t1 = {
            "text": b"hello",
            "hash": hashlib.sha256(b"data1").digest(),
        }
        t2 = {
            "text": b"world", 
            "hash": hashlib.sha256(b"data2").digest(),
        }
        
        diff = DeepDiff(t1, t2)
        
        # Should serialize without errors
        json_output = diff.to_json()
        assert json_output
        
        # Should contain both UTF-8 decoded strings and base64 encoded hashes
        assert "hello" in json_output
        assert "world" in json_output
        
        # Hash values should be base64 encoded
        expected_hash1 = base64.b64encode(t1["hash"]).decode('ascii') 
        expected_hash2 = base64.b64encode(t2["hash"]).decode('ascii')
        assert expected_hash1 in json_output
        assert expected_hash2 in json_output
